@@ -1,4 +1,4 @@
-/* Copyright (c) 2009-2012 Sam Trenholme
+/* Copyright (c) 2009-2014 Sam Trenholme
  *
  * TERMS
  *
@@ -33,14 +33,39 @@ extern int32_t key_n[];
 dns_string *dwc_init_dns_str(int32_t ancount, int32_t nscount,
                 int32_t arcount) {
         dns_string *out = 0;
+        int a;
 
-        out = malloc(sizeof(dns_string));
-        if(out == 0 || ancount < 0 || nscount < 0 || arcount < 0) {
+        out = dw_malloc(sizeof(dns_string));
+        if(out == 0) {
                 return 0;
         }
-        out->an = malloc((ancount * sizeof(uint16_t) * 2) + 2);
-        out->ns = malloc((nscount * sizeof(uint16_t) * 2) + 2);
-        out->ar = malloc((arcount * sizeof(uint16_t) * 2) + 2);
+        if(ancount < 0 || nscount < 0 || arcount < 0) {
+                goto catch_dwc_init_dns_str;
+        }
+        out->an = dw_malloc((ancount * sizeof(uint16_t) * 2) + 2);
+        if(out->an != 0) {
+                for(a = 0; a < (ancount * 2); a++) {
+                        out->an[a] = 0;
+                }
+        } else if(ancount > 0) {
+                goto catch_dwc_init_dns_str;
+        }
+        out->ns = dw_malloc((nscount * sizeof(uint16_t) * 2) + 2);
+        if(out->ns != 0) {
+                for(a = 0; a < (nscount * 2); a++) {
+                        out->ns[a] = 0;
+                }
+        } else if(nscount > 0) {
+                goto catch_dwc_init_dns_str;
+        }
+        out->ar = dw_malloc((arcount * sizeof(uint16_t) * 2) + 2);
+        if(out->ar != 0) {
+                for(a = 0; a < (arcount * 2); a++) {
+                        out->ar[a] = 0;
+                }
+        } else if(arcount > 0) {
+                goto catch_dwc_init_dns_str;
+        }
         out->packet = 0;
         out->ancount = ancount;
         out->nscount = nscount;
@@ -64,6 +89,7 @@ catch_dwc_init_dns_str:
         if(out->ar != 0) {
                 free(out->ar);
         }
+        free(out);
         return 0;
 }
 
