@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2011 Sam Trenholme
+/* Copyright (c) 2007-2012 Sam Trenholme
  *
  * TERMS
  *
@@ -400,6 +400,7 @@ void dwm_init_mararc() {
         key_n[DWM_N_truncation_hack] = 1;
         key_n[DWM_N_reject_ptr] = 0;
         key_n[DWM_N_min_ttl_incomplete_cname] = 3600;
+        key_n[DWM_N_max_ttl] = 86400;
 }
 
 /* Look for a Mararc parameter; -1 if not found/error; 0-n if found
@@ -505,7 +506,7 @@ void dwm_execfile(dw_str *execfile, dw_str *fname) {
  * 3: Set dictionary variable
  * 4: Append to dictionary variable
  * 5: Set numeric variable
- * 6: Init dictionary variable (currently not used)
+ * 6: Init dictionary variable
  * 7: Read other file for more dwood2rc parameters
  */
 
