@@ -71,6 +71,7 @@ int32_t min_bind = 15000;
 int32_t num_ports = 4096;
 int32_t maradns_uid = 99;
 int32_t maradns_gid = 99;
+int32_t max_ttl = 86400;
 int num_retries = 5;
 dwd_dict *blacklist_dict = 0;
 
@@ -500,6 +501,10 @@ void process_numeric_mararc_params() {
         maradns_gid = get_key_n(DWM_N_maradns_gid,10,65535,99);
         resurrections = get_key_n(DWM_N_resurrections,0,1,1);
         num_retries = get_key_n(DWM_N_num_retries,0,32,5);
+        max_ttl = get_key_n(DWM_N_max_ttl,
+                300 /* 5 minutes */,
+                7776000 /* 90 days */,
+                86400 /* One day */);
 
         if((num_ports & (num_ports - 1)) != 0) {
                 dw_fatal("num_ports must be a power of 2");
@@ -547,6 +552,9 @@ int process_root_upstream_servers(int param, int is_upstream, char *bad) {
                 q = dw_copy(r);
                 dw_destroy(r);
         }
+        if(a == 20000) {
+                dw_fatal("Too many upstream servers, limit 20,000");
+        }
 
         return out;
 
@@ -623,8 +631,8 @@ void process_mararc_params() {
 
         /* These two items are allocated once and never freed;
          * they are used while the program is running */
-        b_remote = malloc((maxprocs + 1) * sizeof(int));
-        rem = malloc((maxprocs + 1) * sizeof(remote_T));
+        b_remote = dw_malloc((maxprocs + 1) * sizeof(int));
+        rem = dw_malloc((maxprocs + 1) * sizeof(remote_T));
 
         /* Make sure we got that memory */
         if(b_remote == 0) {
