@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2010 Sam Trenholme
+/* Copyright (c) 2007-2014 Sam Trenholme
  *
  * TERMS
  *
@@ -48,16 +48,17 @@ int dw_assert_sanity(dw_str *object) {
  */
 dw_str *dw_create(uint32_t size) {
         dw_str *new;
-        new = malloc(sizeof(dw_str));
+        new = dw_malloc(sizeof(dw_str));
         if(new == 0) {
                 return 0; /* We should probably give an "aiee" and
                            * exit(1) here */
         }
         if(size >= 134217728 /* 2 ** 27 */) {
+                free(new);
                 return 0;
         }
         /* 2 byte cushion to avoid off-by-one security problems */
-        new->str = malloc(size + 2);
+        new->str = dw_malloc(size + 2);
         if(new->str == 0) {
                 free(new);
                 return 0;
@@ -289,7 +290,7 @@ dw_str *dw_substr(dw_str *obj, int32_t begin, int32_t amount, int32_t max) {
         /* Do the actual copying */
         for(c = 0; c < amount; c++) {
                 /* Bounds checking */
-                if(c > begin + obj->len || c >= copy->max) {
+                if(c + begin > obj->len || c >= copy->max) {
                         break;
                 }
                 *(copy->str + c) = *(obj->str + begin + c);
@@ -582,7 +583,7 @@ uint8_t *dw_to_cstr(dw_str *obj) {
                 return 0;
         }
 
-        out = malloc(obj->len + 3);
+        out = dw_malloc(obj->len + 3);
         for(a = 0; a < obj->len; a++) {
                 *(out + a) = *(obj->str + a);
         }
@@ -662,7 +663,7 @@ dw_str *dw_qspop(dw_str *in) {
         }
         out = dw_substr(in,index,len,1);
         in->len -= len + 1;
-        if(in->len < 0 || in->len >= in->max) {
+        if(in->len >= in->max) {
                 in->len = 0;
                 goto catch_dw_qspop;
         }
@@ -1026,7 +1027,7 @@ dw_str *dw_packet_to_cache(uint8_t *raw, int len, uint8_t type) {
 
         /* A very ugly, but fast, way to make recvfrom()'s str/len pair
          * in to a dw_str object.  */
-        hack = malloc(sizeof(dw_str));
+        hack = dw_malloc(sizeof(dw_str));
         hack->max = len + 1;
         hack->len = len;
         hack->str = (uint8_t *)raw;
