@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2011 Sam Trenholme
+/* Copyright (c) 2007-2014 Sam Trenholme
  *
  * TERMS
  *
@@ -21,6 +21,8 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <time.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #ifndef MINGW
 #include <grp.h>
 #include <signal.h>
@@ -35,6 +37,10 @@ int64_t the_time = 0;
 /* RNG seed/state */
 dwr_rg *rng_seed;
 
+/* Whether to read the cache file */
+int do_read_cache = 1;
+struct stat mararc_st;
+
 /* Mararc parameters that are set in DwMararc.c */
 extern dw_str *key_s[];
 extern dw_str *key_d[];
@@ -359,8 +365,9 @@ void dw_fatal(char *why) {
 
 /* Set the 64-bit timestamp starting at 290805600 unix() time (When
  * the Blake's 7 episode Gambit was originally broadcast); this should
- * be called once a second or so */
+ * be called several times a second.  Each second has 256 "ticks". */
 void set_time() {
+#ifdef FALLBACK_TIME
         time_t sys_time;
         sys_time = time(0);
         if(sizeof(sys_time) > 4) {
@@ -375,6 +382,45 @@ void set_time() {
                 }
         }
         the_time <<= 8; /* Each second has 256 "ticks" */
+#else /* FALLBACK_TIME */
+#ifndef MINGW
+        struct timespec posix_time;
+        time_t coarse;
+        long fine;
+        long result;
+        result = clock_gettime(CLOCK_REALTIME, &posix_time);
+        if(result == 0) { /* Successful getting time */
+                coarse = posix_time.tv_sec;
+                fine = posix_time.tv_nsec;
+                if(sizeof(coarse) > 4) {
+                        if(coarse != -1) {
+                                the_time = coarse - 290805600;
+                        }
+                } else {
+                        if(coarse < DW_MINTIME) {
+                                the_time = coarse + 4004161696U;
+                        } else {
+                                the_time = coarse - 290805600;
+                        }
+                }
+                the_time <<= 8;
+                fine /= 3906250; /* 256 "ticks" per second */
+                if(fine > 0 && fine <= 256) {
+                        the_time += fine;
+                }
+                //printf("time: %llx control %lx\n",the_time,time(0)-290805600);//DEBUG
+        }
+#else /* MINGW */
+        FILETIME win_time = { 0, 0 };
+        GetSystemTimeAsFileTime(&win_time);
+        the_time = win_time.dwHighDateTime & 0xffffffff;
+        the_time <<= 32;
+        the_time |= (win_time.dwLowDateTime & 0xffffffff);
+        the_time *= 2;
+        the_time /= 78125;
+        the_time -= 3055431475200LL;
+#endif /* MINGW */
+#endif /* FALLBACK_TIME */
 }
 
 /* Set up some signal handlers, so MaraDNS can write the cache to a file
@@ -447,6 +493,7 @@ void process_signal(int number) {
 void init_cache() {
         dw_str *filename = 0;
         char *fname_convert = 0;
+        struct stat cache_st;
 
         dwh_process_mararc_params(); /* Get the cache size */
         if(cache != 0) { /* Don't init cache twice */
@@ -455,10 +502,16 @@ void init_cache() {
 
         /* See if we can read the cache from a file */
         filename = key_s[DWM_S_cache_file];
-        if(filename != 0) {
+        if(filename != 0 && do_read_cache == 1) {
                 dw_filename_sanitize(filename);
                 fname_convert = (char *)dw_to_cstr(filename);
-                cache = dwh_read_hash(fname_convert);
+                if(stat(fname_convert,&cache_st) == 0 &&
+                   cache_st.st_mtime < mararc_st.st_mtime) {
+                        dw_log_string(
+                              "Cache older than rc file; not reading cache",0);
+                } else {
+                        cache = dwh_read_hash(fname_convert);
+                }
                 free(fname_convert);
         }
 
@@ -468,7 +521,7 @@ void init_cache() {
 }
 
 /* Read mararc parameters and set global variables based on those
- * parameters */
+ * parameters; disable cache reading if cache file is more recent */
 void process_mararc(char *name) {
         if(dwm_parse_mararc(name) != 1) {
                 dw_log_3strings("Fatal error parsing file ",name,"",1);
@@ -478,6 +531,12 @@ void process_mararc(char *name) {
         /*if(key_s == 0 || key_d == 0) {
                 dw_fatal("error getting mararc parameters");
         }*/
+        /* If the dwood3rc is newer than the cache file, do not read the
+         * cache */
+        if(stat(name,&mararc_st) != 0) {
+                dw_log_string("Can not stat rc file; not reading cache",0);
+                do_read_cache = 0;
+        }
 }
 
 /* Given a C-string string containing random noise, the length of that
@@ -515,7 +574,11 @@ void get_entropy_from_seedfile(uint8_t *noise,int len) {
         int seed = -1;
 
         if(key_s[DWM_S_random_seed_file] == 0) {
+#ifdef MINGW
+                filename = "secret.txt"; /* Default filename */
+#else /* MINGW */
                 filename = "/dev/urandom"; /* Default filename */
+#endif /* MINGW */
         } else {
                 filename = (char *)dw_to_cstr(key_s[DWM_S_random_seed_file]);
                 zap = 1;
@@ -551,16 +614,16 @@ void init_rng() {
         int64_t tstamp = 0;
         pid_t pnum = 1;
 
-        noise = (uint8_t *)malloc(512);
+        noise = (uint8_t *)dw_malloc(512);
+        if(noise == 0) {
+                dw_fatal("error allocating memory for noise");
+        }
 #ifdef VALGRIND_NOERRORS
         /* Valgrind reports our intentional use of values of uncleared
          * allocated memory as one source of entropy as an error, so we
          * allow it to be disabled for Valgrind testing */
         memset(noise,0,512);
 #endif /* VALGRIND_NOERRORS */
-        if(noise == 0) {
-                dw_fatal("error allocating memory for noise");
-        }
 
         get_entropy_from_seedfile(noise,256);
 
@@ -591,7 +654,8 @@ void init_rng() {
 
 /* Drop privileges and become unprivileged user */
 void sandbox() {
-#if ! (defined MINGW || defined __CYGWIN__)
+#ifndef MINGW
+#ifndef __CYGWIN__
         unsigned char *c = 0;
         gid_t g = DW_UID;
         if(key_s[DWM_S_chroot_dir] == 0) {
@@ -602,13 +666,14 @@ void sandbox() {
                 dw_fatal("Converting chroot_dir to string failed");
         }
         if(chdir((char *)c) != 0) {
+                printf("There is no directory %s\n",(char *)c);
                 dw_fatal("chdir() failed");
         }
 #ifndef QNX
         if(chroot((char *)c) == -1) {
                 dw_fatal("chroot() failed");
         }
-#endif
+#endif /* QNX */
         if(setgroups(1,&g) == -1) {
                 dw_fatal("setgroups() failed");
         }
@@ -627,6 +692,7 @@ void sandbox() {
                 c = 0;
         }
         return;
+#endif /* __CYGWIN__ */
 #endif /* MINGW */
 }
 
