@@ -1,4 +1,4 @@
-/* Copyright (c) 2007-2012 Sam Trenholme and others
+/* Copyright (c) 2007-2014 Sam Trenholme and others
  *
  * TERMS
  *
@@ -56,6 +56,7 @@ extern int resurrections;
 extern int min_bind;
 extern int num_ports;
 extern int num_retries;
+extern int32_t max_ttl;
 
 /* Other mararc parameters */
 extern dwd_dict *blacklist_dict;
@@ -100,7 +101,7 @@ void setup_bind(sockaddr_all_T *dns_udp, uint16_t port, int len) {
         if(dns_udp == 0) {
                 return;
         }
-        memset(dns_udp,0,sizeof(dns_udp));
+        memset(dns_udp,0,sizeof(*dns_udp));
         if(len == 4) {
                 dns_udp->V4.sin_family = AF_INET;
                 dns_udp->V4.sin_addr.s_addr = htonl(INADDR_ANY);
@@ -154,7 +155,7 @@ SOCKET setup_server(sockaddr_all_T *server, ip_addr_T *addr) {
                 return INVALID_SOCKET;
         }
 
-        memset(server,0,sizeof(server));
+        memset(server,0,sizeof(*server));
         if ( addr->len == 4 ) {
                 server->V4.sin_family = AF_INET;
                 server->V4.sin_port = htons(upstream_port);
@@ -333,7 +334,7 @@ void make_new_udp_connect(int b, unsigned char *a, int len, int num_alloc) {
         }
         rem[b].query = dw_get_dname_type(a,12,len);
         dwc_lower_case(rem[b].query);
-        rem[b].local = malloc(num_alloc * sizeof(local_T *));
+        rem[b].local = dw_malloc(num_alloc * sizeof(local_T *));
         if(rem[b].local == 0) {
                 return;
         }
@@ -394,7 +395,7 @@ int forward_local_udp_packet(SOCKET sock, int32_t local_id,
 #endif /* INFLIGHT_VERBOSE */
         }
 
-        rem[b].local[rem[b].num_locals - 1] = malloc(sizeof(local_T));
+        rem[b].local[rem[b].num_locals - 1] = dw_malloc(sizeof(local_T));
         if(rem[b].local[rem[b].num_locals - 1] != 0) {
                 rem[b].local[rem[b].num_locals - 1]->orig_query = 0;
                 rem[b].local[rem[b].num_locals - 1]->action = 0;
@@ -529,6 +530,10 @@ int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
         socklen_t c_len = sizeof(struct sockaddr_in);
         int ret = -1, type = 0, cache_type = 0;
 
+        if(client == 0) {
+                goto catch_get_reply_from_cache;
+        }
+
 #ifdef IPV6
         if (client->Family == AF_INET6) {
                c_len = sizeof(struct sockaddr_in6);
@@ -813,6 +818,9 @@ int cache_dns_reply(unsigned char *packet, int count, int b, int truncated) {
         } else {
                 answer = dw_packet_to_cache(packet,count,is_nxdomain);
                 decomp = dwc_decompress(question,answer);
+                if(decomp == 0) {
+                        goto catch_cache_dns_reply;
+                }
                 if(dwc_has_bad_ip(decomp,blacklist_dict)) {
                         ret = -2; /* Tell caller we need synth "not there" */
                         goto catch_cache_dns_reply;
@@ -836,6 +844,9 @@ int cache_dns_reply(unsigned char *packet, int count, int b, int truncated) {
                 if(ttl < 30) {
                         ttl = 30;
                 }
+                if(ttl > max_ttl) {
+                        ttl = max_ttl;
+                }
 
                 /* Routines in DwRecurse.c process the packet and let us know
                  * what kind of packet we got upstream (so we know how to
@@ -905,7 +916,7 @@ unsigned char *make_synth_not_there_answer(unsigned char *a, int *count,
         unsigned char *answer = 0;
         int counter = 0;
 
-        answer = malloc(*count + 43);
+        answer = dw_malloc(*count + 43);
         if(answer == 0) {
                 return 0;
         }
@@ -1162,6 +1173,24 @@ int get_rem_udp_packet_core(unsigned char *a, ssize_t count,
                 return 2; /* Don't kill pending UDP connection */
         } else { /* Truncated over UDP; just given them a blank "truncated"
                   * reply */
+                if(key_n[DWM_N_tcp_listen] != 1) {
+                        /* EasyDNS sometimes has given out packets marked
+                         * "truncated" that, in violation of RFC1035 section
+                         * 4.1.1, do not mean that "[the] message was
+                         * truncated due to length greater than that
+                         * permitted on the transmission channel.", but mean
+                         * "our UDP server is broken, try using our TCP
+                         * server".
+                         *
+                         * This in mind, if we got a truncated packet and
+                         * can not extract any useful information from the
+                         * packet, unless Deadwood is using DNS-over-TCP,
+                         * it's better to completely ignore the reply (when
+                         * EasyDNS has had this issue, only some of their
+                         * DNS servers have been affected)
+                         */
+                        return -1;
+                }
                 cache_dns_reply(a,count,b,1);
         }
 
@@ -1192,6 +1221,17 @@ void get_remote_udp_packet(int b, SOCKET sock) {
         }
         a[2] |= 0x80; /* Flag this as an answer (just in case they didn't) */
 
+#ifdef SHOWPACKET
+        dw_str *hack = 0;
+        hack = dw_malloc(sizeof(dw_str));
+        hack->max = count + 1;
+        hack->len = count;
+        hack->str = (uint8_t *)a;
+        hack->sane = 114;
+        dw_log_dwstr("Packet received ",hack,0);
+        free(hack);
+#endif /* SHOWPACKET */
+
         /* Make reasonably sure this is the reply to their question */
         if(verify_dns_packet(b,a,count) != 1) {
                 return;
