@@ -1,4 +1,4 @@
-/* Copyright (c) 2002 Sam Trenholme
+/* Copyright (c) 2002-2006 Sam Trenholme
  *
  * TERMS
  *
@@ -20,6 +20,7 @@
    data to send over the network */
 
 #include "../MaraDns.h"
+#include "functions_dns.h"
 
 /* Determine the length of a domain-name label (if it is a compressed
    label, determine the length of the compressed portion)
@@ -47,9 +48,9 @@ int dlabel_length(js_string *raw, unsigned int offset) {
     while(length < 256 && toread > 0 && toread != '_') {
         if(toread >= 192) /* compressed label */ {
             if(length + 2 + offset <= raw->unit_count)
-	       return length + 2;
+               return length + 2;
             else
-	       return JS_ERROR;
+               return JS_ERROR;
             }
         if(toread > 63)
             return JS_ERROR; /* No EDNS support */
@@ -87,9 +88,9 @@ void init_header(q_header *header) {
     }
 
 /* Given a q_header structure, and a js object to put the raw UDP data
-   in, convert a q_header structure to raw UDP data. 
+   in, convert a q_header structure to raw UDP data.
    input: pointer to q_header structure, pointer to js_string object
-   output: JS_ERROR on error, JS_SUCCESS on success 
+   output: JS_ERROR on error, JS_SUCCESS on success
 */
 
 int make_hdr(q_header *header, js_string *js) {
@@ -99,7 +100,7 @@ int make_hdr(q_header *header, js_string *js) {
     if(js_has_sanity(js) == JS_ERROR)
        return JS_ERROR;
     if(js->unit_size != 1)
-        return JS_ERROR; 
+        return JS_ERROR;
     if(js->max_count < 14)
         return JS_ERROR;
 
@@ -111,7 +112,7 @@ int make_hdr(q_header *header, js_string *js) {
     raw[1] = (header->id & 0x00ff);
     /* All the flags and 4-bit fields are a little tricky */
     raw[2] = raw[3] = 0;
-    if(header->qr == 1) 
+    if(header->qr == 1)
         raw[2] |= 0x80;
     raw[2] |= (header->opcode & 0x0f) << 3;
     if(header->aa == 1)
@@ -145,7 +146,7 @@ int make_hdr(q_header *header, js_string *js) {
 /* Given a q_header structure, and a js object with the raw UDP data,
    convert raw UDP data to a q_header structure.
    input: pointer to q_header structure, pointer to js_string object
-   output: JS_ERROR on error, JS_SUCCESS on success 
+   output: JS_ERROR on error, JS_SUCCESS on success
 */
 
 int read_hdr(js_string *js, q_header *header) {
@@ -155,7 +156,7 @@ int read_hdr(js_string *js, q_header *header) {
     if(js_has_sanity(js) == JS_ERROR)
        return JS_ERROR;
     if(js->unit_size != 1)
-        return JS_ERROR; 
+        return JS_ERROR;
     if(js->max_count < 14)
         return JS_ERROR;
 
@@ -198,7 +199,7 @@ int read_hdr(js_string *js, q_header *header) {
     }
 
 /* Given a js string object and a q_question structure, place the raw UDP
-   format of the query at the end of the js_string object 
+   format of the query at the end of the js_string object
    input: see above
    output: JS_ERROR on error, JS_SUCCESS on success */
 
@@ -209,8 +210,8 @@ int make_question(q_question *question, js_string *js) {
 
     int offset = js->unit_count;
 
-    char *raw;
-     
+    unsigned char *raw;
+
     counter = toread = 0;
 
     /* Some sanity checks */
@@ -222,7 +223,7 @@ int make_question(q_question *question, js_string *js) {
         return JS_ERROR;
     if(question->qname->unit_size != 1)
         return JS_ERROR;
-    if(offset >= js->max_count) 
+    if(offset >= js->max_count)
         return JS_ERROR;
 
     raw = question->qname->string;
@@ -232,20 +233,20 @@ int make_question(q_question *question, js_string *js) {
         toread = *(raw + counter);
         if(toread > 63) /* To do: compression support */
             return JS_ERROR;
-	read = 0;
-	if(toread == 0) /* 0-length query means we are at the dot after .com */
-	    break;
-	while(read <= toread) {
-	    /* Overflow protection */
-	    if(counter < question->qname->max_count && 
-	       counter + offset < js->max_count) 
-	        /* Then copy the character over */
-		*(js->string + offset + counter) = *(raw + counter);
+        read = 0;
+        if(toread == 0) /* 0-length query means we are at the dot after .com */
+            break;
+        while(read <= toread) {
+            /* Overflow protection */
+            if(counter < question->qname->max_count &&
+               counter + offset < js->max_count)
+                /* Then copy the character over */
+                *(js->string + offset + counter) = *(raw + counter);
             else
-	        return JS_ERROR;
-	    read++;
-	    counter++;
-	    }
+                return JS_ERROR;
+            read++;
+            counter++;
+            }
         }
 
     /* Since counter is now sitting on the dot after .com, increment it */
@@ -255,11 +256,11 @@ int make_question(q_question *question, js_string *js) {
     /* Add the QTYPE and QCLASS to the raw UDP data */
     if(counter + offset + 4 < js->max_count) {
         raw = js->string + counter + offset;
-	raw[0] = (question->qtype & 0xff00) >> 8;
-	raw[1] = question->qtype & 0x00ff;
-	raw[2] = (question->qclass & 0xff00) >> 8;
-	raw[3] = question->qclass & 0x00ff;
-	}
+        raw[0] = (question->qtype & 0xff00) >> 8;
+        raw[1] = question->qtype & 0x00ff;
+        raw[2] = (question->qclass & 0xff00) >> 8;
+        raw[3] = question->qclass & 0x00ff;
+        }
     else
         return JS_ERROR;
 
@@ -282,8 +283,8 @@ int read_question(js_string *js,q_question *question, int offset) {
 
     unsigned char toread, read;
 
-    char *raw;
-     
+    unsigned char *raw;
+
     counter = toread = 0;
 
     /* Some sanity checks */
@@ -295,7 +296,7 @@ int read_question(js_string *js,q_question *question, int offset) {
         return JS_ERROR;
     if(question->qname->unit_size != 1)
         return JS_ERROR;
-    if(offset >= js->unit_count) 
+    if(offset >= js->unit_count)
         return JS_ERROR;
 
     raw = js->string + offset;
@@ -305,21 +306,21 @@ int read_question(js_string *js,q_question *question, int offset) {
         toread = *(raw + counter);
         if(toread > 63) /* To do: compression support */
             return JS_ERROR;
-	read = 0;
-	if(toread == 0) /* 0-length query means we are at the dot after .com */
-	    break;
-	while(read <= toread) {
-	    /* Overflow protection */
-	    if(counter < question->qname->max_count && 
-	       counter + offset < js->unit_count) 
-	        /* Then copy the character over */
-		*(question->qname->string + counter) = *(raw + counter);
+        read = 0;
+        if(toread == 0) /* 0-length query means we are at the dot after .com */
+            break;
+        while(read <= toread) {
+            /* Overflow protection */
+            if(counter < question->qname->max_count &&
+               counter + offset < js->unit_count)
+                /* Then copy the character over */
+                *(question->qname->string + counter) = *(raw + counter);
             else
-	        return JS_ERROR;
-	    read++;
-	    counter++;
-	    }
-         
+                return JS_ERROR;
+            read++;
+            counter++;
+            }
+
         }
 
     /* Since counter is now sitting on the dot after .com, increment it */
@@ -329,9 +330,9 @@ int read_question(js_string *js,q_question *question, int offset) {
     /* Get the QTYPE and QCLASS from the raw UDP data */
     if(counter + offset + 4 <= js->unit_count) {
         raw = js->string + counter + offset;
-	question->qtype = (raw[0] << 8 & 0xff00) | (raw[1] & 0xff);
-	question->qclass = (raw[2] << 8 & 0xff00) | (raw[3] & 0xff);
-	}
+        question->qtype = (raw[0] << 8 & 0xff00) | (raw[1] & 0xff);
+        question->qclass = (raw[2] << 8 & 0xff00) | (raw[3] & 0xff);
+        }
     else
         return JS_ERROR;
 
@@ -353,9 +354,10 @@ int read_question(js_string *js,q_question *question, int offset) {
 */
 
 int hname_translate(js_string *hostname, int qtype) {
-    int counter, special_root_case = 0;
-    unsigned char toread;
-    char *raw;
+    int counter = 0, special_root_case = 0;
+    unsigned char toread = 0;
+    unsigned char *raw = 0;
+    int dotseen = 0;
 
     /* Sanity checks */
     if(js_has_sanity(hostname) == JS_ERROR)
@@ -369,70 +371,82 @@ int hname_translate(js_string *hostname, int qtype) {
     toread = *(hostname->string);
 
     /* Handle the special case of a zero-length zone name */
-    if(hostname->unit_count == 1 && toread == 0) 
+    if(hostname->unit_count == 1 && toread == 0)
         special_root_case = 1;
 
     /* Change the first letter based on the Query type */
     raw = hostname->string;
     switch(qtype) {
         case RR_A:
-	    *raw = 'A';
-	    break;
+            *raw = 'A';
+            break;
         case RR_NS:
-	    *raw = 'N';
+            *raw = 'N';
             break;
         case RR_CNAME:
-	    *raw = 'C';
+            *raw = 'C';
             break;
         case RR_SOA:
-	    *raw = 'S';
-	    break;
+            *raw = 'S';
+            break;
         case RR_PTR:
-	    *raw = 'P';
+            *raw = 'P';
+            break;
+        case RR_MX:
+            *raw = '@';
             break;
-	case RR_MX: 
-	    *raw = '@';
-	    break;
         case RR_TXT:
-	    *raw = 'T';
-	    break;
+            *raw = 'T';
+            break;
         case RR_ANY:
             *raw = 'Z';
-	    break;
+            break;
         case -2: /* Special command to make it a pipe */
-	    *raw = '|';
-	    break;
+            *raw = '|';
+            break;
+        case RR_MAGIC_SPACE: /* Special command to make it a space */
+        case RR_MAGIC_EMAIL: /* Same as magic space, but make the first
+                              * seperator an @ instead of a dot */
+            *raw = ' ';
+            break;
         default:
-	    *raw = 'U';
+            *raw = 'U';
         }
 
      /* Again, handle the special case of a zero-length host name
         (the "root" of the DNS tree) */
      if(special_root_case == 1) {
         if(hostname->max_count < 2)
-	    return JS_ERROR;
+            return JS_ERROR;
         hostname->unit_count = 2;
-	raw++;
-	*raw = '.';
-	return JS_SUCCESS;
-	}
-        
+        raw++;
+        *raw = '.';
+        return JS_SUCCESS;
+        }
+
      /* Convert the hostname delimters in to dots */
      counter = 0;
+     dotseen = 0;
      while(counter < 256 && toread > 0) {
          if(toread > 63)
-	     return JS_ERROR; /* To do: compression support */
+             return JS_ERROR; /* To do: compression support */
          counter += toread + 1;
-	 /* Get the next "jump" we need to do */
-         if(counter > hostname->unit_count) 
-             return JS_ERROR; 
-	 toread = *(raw + counter);
-	 /* Buffer overflow protection */
-	 if(counter <= hostname->unit_count)
-	     *(raw + counter) = '.';
-         else
-	     return JS_ERROR;
-         } 
+         /* Get the next "jump" we need to do */
+         if(counter > hostname->unit_count)
+             return JS_ERROR;
+         toread = *(raw + counter);
+         /* Buffer overflow protection */
+         if(counter <= hostname->unit_count) {
+             if(qtype != RR_MAGIC_EMAIL || dotseen > 0) {
+                 *(raw + counter) = '.';
+             } else {
+                 *(raw + counter) = '@';
+                 dotseen = 1;
+             }
+         } else {
+             return JS_ERROR;
+             }
+         }
 
      return JS_SUCCESS;
      }
@@ -447,7 +461,7 @@ int email_translate(js_string *hostname) {
     int counter;
     unsigned char toread;
     int first = 1;
-    char *raw;
+    unsigned char *raw;
 
     /* Sanity checks */
     if(js_has_sanity(hostname) == JS_ERROR)
@@ -468,38 +482,45 @@ int email_translate(js_string *hostname) {
      counter = 0;
      while(counter < 256 && toread > 0) {
          if(toread > 63)
-	     return JS_ERROR; /* To do: compression support */
+             return JS_ERROR; /* To do: compression support */
          counter += toread + 1;
-	 /* Get the next "jump" we need to do */
-         if(counter > hostname->unit_count) 
-             return JS_ERROR; 
-	 toread = *(raw + counter);
-	 /* Buffer overflow protection */
-	 if(counter <= hostname->unit_count) {
+         /* Get the next "jump" we need to do */
+         if(counter > hostname->unit_count)
+             return JS_ERROR;
+         toread = *(raw + counter);
+         /* Buffer overflow protection */
+         if(counter <= hostname->unit_count) {
              if(first) {
-	         *(raw + counter) = '@';
+                 *(raw + counter) = '@';
                  first = 0;
                  }
              else {
-	         *(raw + counter) = '.';
+                 *(raw + counter) = '.';
                  }
              }
          else
-	     return JS_ERROR;
-         } 
+             return JS_ERROR;
+         }
 
      return JS_SUCCESS;
      }
 
 /* Zero-copy implemention of converting a human-readable host name in to the
-   raw RFC1035 UDP data for a domain. 
+   raw RFC1035 UDP data for a domain.
    Input:  pointer to js_string object we modify "in place"
-   Output: -1 on error, numeric type of query on success, -2 on 
+   Output: -1 on error, numeric type of query on success, -2 on
            unsupported query type, -3 on 'U' query type (which
            then has to be specified by the user elsewhere)
 */
 
 int hname_2rfc1035(js_string *hostname) {
+    return hname_2rfc1035_starwhitis(hostname,0);
+    }
+
+/* Starwhitis: 0: Stars *not* allowed at end of host name
+               1: Stars are allowed at host name ends */
+
+int hname_2rfc1035_starwhitis(js_string *hostname, int starwhitis) {
     int counter,seen;
     unsigned char *raw, *towrite;
     int ret = -2;
@@ -519,7 +540,7 @@ int hname_2rfc1035(js_string *hostname) {
     switch(*raw) {
         case 'A':
             ret = RR_A;
-	    break;
+            break;
         case 'N':
             ret = RR_NS;
             break;
@@ -527,65 +548,80 @@ int hname_2rfc1035(js_string *hostname) {
             ret = RR_CNAME;
             break;
         case 'S':
-	    ret = RR_SOA;
-	    break;
+            ret = RR_SOA;
+            break;
         case 'P':
-	    ret = RR_PTR;
+            ret = RR_PTR;
+            break;
+        case '@':
+            ret = RR_MX;
             break;
-	case '@': 
-	    ret = RR_MX;
-	    break;
         case 'T':
-	    ret = RR_TXT;
-	    break;
+            ret = RR_TXT;
+            break;
         case 'U':
             ret = -3;
             break;
         case 'Z':
-	    ret = RR_ANY;
+            ret = RR_ANY;
             break;
         default:
-	    return -2;
+            return -2;
         }
 
      /* Handle the trivial case of an input in the form "A." */
      if(*(raw + 1) == '.' && hostname->unit_count == 2) {
          *raw = 0;
          hostname->unit_count = 1;
-	 return ret;
+         return ret;
+         }
+
+     /* Also handle the case of the hostname being *just* a star */
+     if(*(raw + 1) == '*' && hostname->unit_count == 2 &&
+                     starwhitis == 1) {
+         *raw = '_';
+         hostname->unit_count = 1;
+         return ret;
          }
 
      /* Convert the dots in to hostname delimiters */
      seen = counter = 0;
      while(counter < 256 && counter < hostname->unit_count) {
          counter++;
-         if(counter > hostname->unit_count) 
-             return JS_ERROR; 
+         if(counter > hostname->unit_count)
+             return JS_ERROR;
          if(*(raw + counter) == '.') {
-	     if(seen < 1 || seen > 63)
-	         return JS_ERROR;
-	     *towrite = seen;
-	     towrite = raw + counter;
-	     seen = 0;
-	     }
+             if(seen < 1 || seen > 63)
+                 return JS_ERROR;
+             *towrite = seen;
+             towrite = raw + counter;
+             seen = 0;
+             }
          else
-	     seen++;
+             seen++;
          if(counter == hostname->unit_count - 1) /* The final '.' */
              break;
-         } 
+         }
 
      /* That last '0' for the dot after .com */
-     if(counter < 256 && *(raw + counter) == '.')
+     if(counter < 256 && *(raw + counter) == '.') {
          *(raw + counter) = 0;
-     else 
+     /* We need to deal with trailing * characters */
+     } else if(counter < 256 && counter > 1 && *(raw + counter) == '*' &&
+                     hostname->unit_count > 1 && starwhitis == 1) {
+         *(raw + counter - 1) = '_';
+         hostname->unit_count--;
+     }
+     else {
          return JS_ERROR;
+     }
 
      return ret;
      }
 
-/* Zero-copy implemention of converting a human-readable email address in to 
-   the raw RFC1035 UDP data for a domain. 
-   Input:  pointer to js_string object we modify "in place" (needs a 
+/* Zero-copy implemention of converting a human-readable email address in to
+   the raw RFC1035 UDP data for a domain.
+   Input:  pointer to js_string object we modify "in place" (needs a
            one-octet 'prefix' which can be be any character
    Output: -1 on error, JS_SUCCESS on success
 */
@@ -609,46 +645,59 @@ int email_2rfc1035(js_string *hostname) {
 
      /* Convert the at/dots in to hostname delimiters */
      seen = counter = 0;
+     /* Handle the special case of this being just a dot by itself */
+     if(hostname->unit_count == 2 && *(raw + 1) == '.') {
+         hostname->unit_count = 1;
+         *(hostname->string) = '\0';
+         return JS_SUCCESS;
+         }
+
      while(counter < 256 && counter < hostname->unit_count) {
          counter++;
-         if(counter > hostname->unit_count) 
-             return JS_ERROR; 
+         if(counter > hostname->unit_count)
+             return JS_ERROR;
          if((!firstat && *(raw + counter) == '.') ||
              (firstat && *(raw + counter) == '@')) {
              firstat = 0;
-	     if(seen < 1 || seen > 63)
-	         return JS_ERROR;
-	     *towrite = seen;
-	     towrite = raw + counter;
-	     seen = 0;
-	     }
+             if(seen < 1 || seen > 63)
+                 return JS_ERROR;
+             *towrite = seen;
+             towrite = raw + counter;
+             seen = 0;
+             }
          else
-	     seen++;
+             seen++;
          if(counter == hostname->unit_count - 1) /* The final '.' */
              break;
-         } 
+         }
+
+     /* Handle the case of this string not having an @ in it by
+      * processing the string as one with just dots */
+     if(firstat) {
+         return hname_2rfc1035_starwhitis(hostname,0);
+         }
 
      /* That last '0' for the dot after .com */
      if(counter < 256 && *(raw + counter) == '.')
          *(raw + counter) = 0;
-     else 
+     else
          return JS_ERROR;
 
      return JS_SUCCESS;
      }
 
-/* Process the header of a RR record as described in section 4.1.3 of 
+/* Process the header of a RR record as described in section 4.1.3 of
    RFC1035.  This converts the contents of a RFC1035 header in to an
    q_rr structure.
    input: js_string obejct with the raw UDP data, q_rr struct to put data
-          in, offset form beginning of string to look at data for 
+          in, offset form beginning of string to look at data for
    output: number of bytes in rr header on success, JS_ERROR on error
 */
 
 int read_rr_h (js_string *js, q_rr *hdr, int offset) {
     int counter;
     unsigned char toread, read;
-    
+
     unsigned char *raw;
 
     counter = toread = 0;
@@ -662,7 +711,7 @@ int read_rr_h (js_string *js, q_rr *hdr, int offset) {
         return JS_ERROR;
     if(hdr->name->unit_size != 1)
         return JS_ERROR;
-    if(offset >= js->max_count) 
+    if(offset >= js->max_count)
         return JS_ERROR;
 
     raw = js->string + offset;
@@ -670,17 +719,17 @@ int read_rr_h (js_string *js, q_rr *hdr, int offset) {
     /* Copy over the domain-name string inside the js_string object */
     while(counter < 256) {
         if(offset + counter > js->unit_count)
-	    return JS_ERROR;
+            return JS_ERROR;
         toread = *(raw + counter);
-        if(toread > 63) /* To do: compression support */ 
+        if(toread > 63) /* To do: compression support */
             return JS_ERROR;
         read = 0;
         if(toread == 0) /* 0-length query means we are at the dot after .com */
             break;
         while(read <= toread) {
             /* Overflow protection */
-            if(counter < hdr->name->max_count && 
-               counter + offset < js->unit_count) 
+            if(counter < hdr->name->max_count &&
+               counter + offset < js->unit_count)
                 /* Then copy the character over */
                 *(hdr->name->string + counter) = *(raw + counter);
             else
@@ -688,7 +737,7 @@ int read_rr_h (js_string *js, q_rr *hdr, int offset) {
             read++;
             counter++;
             }
-         
+
         }
 
     /* Do not forget to copy over that dot after .com */
@@ -699,17 +748,17 @@ int read_rr_h (js_string *js, q_rr *hdr, int offset) {
     /* Get TYPE, CLASS, TTL, and RDLENGTH from the header */
     if(counter + offset + 10 <= js->unit_count) {
         raw = js->string + counter + offset;
-	hdr->type = raw[0] << 8 | raw[1];
-	hdr->class = raw[2] << 8 | raw[3];
-	hdr->ttl = raw[4] << 24 | raw[5] << 16 | raw[6] << 8 | raw[7];
+        hdr->type = raw[0] << 8 | raw[1];
+        hdr->class = raw[2] << 8 | raw[3];
+        hdr->ttl = raw[4] << 24 | raw[5] << 16 | raw[6] << 8 | raw[7];
         hdr->rdlength = raw[8] << 8 | raw[9];
-	}
+        }
     else
         return JS_ERROR;
 
     if(counter < hdr->name->max_count)
         hdr->name->unit_count = counter;
-    else 
+    else
         return JS_ERROR;
 
     return counter + 10;
@@ -726,7 +775,7 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
     unsigned char read,toread;
 
     unsigned char *raw;
-   
+
     /* Sanity checks */
     if(js_has_sanity(js) == JS_ERROR)
         return JS_ERROR;
@@ -756,8 +805,8 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
             break;
         while(read <= toread) {
             /* Overflow protection */
-            if(counter < soa->mname->max_count && 
-               counter + offset < js->unit_count) 
+            if(counter < soa->mname->max_count &&
+               counter + offset < js->unit_count)
                 /* Then copy the character over */
                 *(soa->mname->string + counter) = *(raw + counter);
             else
@@ -765,7 +814,7 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
             read++;
             counter++;
             }
-         
+
         }
 
     *(soa->mname->string + counter) = 0;
@@ -788,8 +837,8 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
             break;
         while(read <= toread) {
             /* Overflow protection */
-            if(counter < soa->rname->max_count && 
-               counter + offset < js->unit_count) 
+            if(counter < soa->rname->max_count &&
+               counter + offset < js->unit_count)
                 /* Then copy the character over */
                 *(soa->rname->string + counter) = *(raw + counter);
             else
@@ -797,7 +846,7 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
             read++;
             counter++;
             }
-         
+
         }
 
     *(soa->rname->string + counter) = 0;
@@ -805,7 +854,7 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
     if(soa->rname->max_count < counter)
         return JS_ERROR;
     soa->rname->unit_count = counter;
-    
+
     if(counter + offset + 20 > js->unit_count)
         return JS_ERROR;
 
@@ -822,8 +871,8 @@ int read_soa(js_string *js, rr_soa *soa, int offset) {
     }
 
 /* read_ns: Read a NS (or any other <domain-name>) record
-   input: js_string object with raw UDP data, js_string object to have just 
-          the NS record, offset from beginning of raw UDP data to get RR 
+   input: js_string object with raw UDP data, js_string object to have just
+          the NS record, offset from beginning of raw UDP data to get RR
    output: JS_ERROR on ERROR, bytes in <domain-name> on success
 */
 
@@ -831,7 +880,7 @@ int read_ns(js_string *in, js_string *out, int offset) {
     uint16 toread,read;
     int counter;
     unsigned char *raw;
-    
+
     /* Sanity checks */
     if(js_has_sanity(in) == JS_ERROR)
         return JS_ERROR;
@@ -856,8 +905,8 @@ int read_ns(js_string *in, js_string *out, int offset) {
             break;
         while(read <= toread) {
             /* Overflow protection */
-            if(counter < out->max_count && 
-               counter + offset < in->unit_count) 
+            if(counter < out->max_count &&
+               counter + offset < in->unit_count)
                 /* Then copy the character over */
                 *(out->string + counter) = *(raw + counter);
             else
@@ -865,7 +914,7 @@ int read_ns(js_string *in, js_string *out, int offset) {
             read++;
             counter++;
             }
-         
+
         }
 
     *(out->string + counter) = 0;
@@ -878,7 +927,7 @@ int read_ns(js_string *in, js_string *out, int offset) {
     return counter;
     }
 
-/* Process the RR portion of a TXT record.  
+/* Process the RR portion of a TXT record.
    Input: pointer to string of uncompressed UDP data, pointer of string to
           put txt record in, offset from beginning of UDP data
    Output: JS_ERROR on error, byes in TXT record on success
@@ -911,10 +960,132 @@ int read_txt(js_string *in, js_string *out, int offset) {
 
     /* Copy the TXT string over */
     offset++;
-    for(counter = 0;counter < toread; counter++) 
+    for(counter = 0;counter < toread; counter++)
         *(out->string + counter) = *(in->string + offset + counter);
 
     out->unit_count = toread;
     return out->unit_count;
     }
 
+/* This function is designed to make a packet which compresses to a size
+ * greater than 512 bytes smaller.  The input is an uncompressed DNS packet;
+ * the output is the same packet with the last DNS record removed from the
+ * packet.  As per RFC2181 9, if we are removing a DNS record from the NS
+ * or AN section of a reply, the TC bit is not set.  Otherwise (as a
+ * reasonable interpretation of the wishes in RFC2181 9), we remove all DNS
+ * information except the header and mark the DNS packet truncated.
+ *
+ * Input: A js_string object with a full uncompressed raw DNS packet.
+ *
+ * Output: A point to the above js_string object on success; 0 on error */
+
+js_string *squeeze_to_fit(js_string *packet) {
+        q_header hdr;
+        int qd,an,ns,ar; /* counts of the various record types */
+        int qc,count;
+        int offset;
+
+        if(read_hdr(packet,&hdr) == JS_ERROR) {
+                return 0;
+        }
+
+        qd = hdr.qdcount;
+        an = hdr.ancount;
+        ns = hdr.nscount;
+        ar = hdr.arcount;
+
+        /* If the packet is already marked truncated, we remove all records
+         * past the record and otherwise return it as is;
+         * In the case of there being no ns nor ar records, we do the same
+         * thing */
+
+        if(hdr.tc == 1 || (ns == 0 && ar == 0)) {
+                hdr.tc = 1;
+                hdr.qdcount = hdr.ancount = hdr.nscount = hdr.arcount = 0;
+                make_hdr(&hdr,packet);
+                return packet;
+        }
+
+        qc = qd;
+        count = an + ns + ar;
+        offset = 12; /* Beginning of first question */
+        /* Jump past all of the questions */
+        while(qc > 0) {
+                int len;
+                len = dlabel_length(packet,offset);
+                if(len == JS_ERROR) {
+                        hdr.tc = 1;
+                        hdr.qdcount = hdr.ancount = hdr.nscount =
+                                hdr.arcount = 0;
+                        make_hdr(&hdr,packet);
+                        return packet;
+                }
+                len += 4; /* Type and class */
+                offset += len;
+                if(offset >= packet->unit_count) {
+                        hdr.tc = 1;
+                        hdr.qdcount = hdr.ancount = hdr.nscount =
+                                hdr.arcount = 0;
+                        make_hdr(&hdr,packet);
+                        return packet;
+                }
+                qc--;
+        }
+
+        /* Jump past all of the answers except for the last one */
+        while(count > 1) {
+                int len;
+                len = dlabel_length(packet,offset);
+                if(len == JS_ERROR) {
+                        hdr.tc = 1;
+                        hdr.qdcount = hdr.ancount = hdr.nscount =
+                                hdr.arcount = 0;
+                        make_hdr(&hdr,packet);
+                        return packet;
+                }
+                len += 8; /* type, class, TTL, and RDLENGTH */
+                offset += len;
+                if(offset + 2 >= packet->unit_count) {
+                        hdr.tc = 1;
+                        hdr.qdcount = hdr.ancount = hdr.nscount =
+                                hdr.arcount = 0;
+                        make_hdr(&hdr,packet);
+                        return packet;
+                }
+                /* Get len from the rdlength of the data (this is why the
+                 * packet *must* be uncompressed) */
+                len = (((*(packet->string + offset ) & 0xff) << 8) +
+                      *(packet->string + offset + 1));
+                offset += len + 2;
+                if(offset >= packet->unit_count) {
+                        hdr.tc = 1;
+                        hdr.qdcount = hdr.ancount = hdr.nscount =
+                                hdr.arcount = 0;
+                        make_hdr(&hdr,packet);
+                        return packet;
+                }
+                count--;
+        }
+
+        /* Now that we've found the last DNS data, remove it from the
+         * packet */
+        if(ar > 0) {
+                ar--;
+                hdr.arcount = ar;
+        } else if(ns > 0) {
+                ns--;
+                hdr.nscount = ns;
+        } else {
+                hdr.tc = 1;
+                hdr.qdcount = hdr.ancount = hdr.nscount = hdr.arcount = 0;
+                make_hdr(&hdr,packet);
+                return packet;
+        }
+
+        hdr.tc = 0;
+        make_hdr(&hdr,packet);
+        packet->unit_count = offset;
+        return packet;
+
+}
+
