@@ -1,4 +1,4 @@
-/* Copyright (c) 2002,2003 Sam Trenholme
+/* Copyright (c) 2002-2009 Sam Trenholme
  *
  * TERMS
  *
@@ -16,35 +16,37 @@
  * fitness for purpose.
  */
 
-
 /* Parse a mararc file */
 
 #include "../MaraDns.h"
 #include "../libs/MaraHash.h"
 #include "ParseMaraRc_en.h"
 #include <stdlib.h>
+#ifndef MINGW32
 #include <pwd.h>
+#endif
 #include <sys/types.h>
 #include <stdio.h>
-
-extern int js_an_chars();
-extern int js_numbers();
-extern int js_space_chars();
+#include <string.h>
 
 /* Keywords that are non-dictionary strings in Mara's rc file */
 
-#define KEYCOUNT 32
+#define KEYCOUNT 55
 
 char *keywords[KEYCOUNT] = {
-        "bind_address", /* Address to bind to (0.0.0.0 is all addresses) */
+        "bind_address", /* IPv4 Addresses to bind to (old name) */
+        "ipv4_bind_addresses", /* IPv4 Addresses to bind to (new name) */
+        "ipv6_bind_address", /* IPv6 Address (singular) to bind to */
+        "csv2_synthip_list", /* List of IPs we synthesize when not supplied in
+                              * csv2 zone files */
         "chroot_dir",   /* Directory to chroot to */
-	"debug_response_delay", /* Seconds to delay in sending a reply
-	                           (useful for debugging purposes) */
+        "debug_response_delay", /* Seconds to delay in sending a reply
+                                   (useful for debugging purposes) */
         "debug_msg_level", /* The level of debug messages to allow.
-	                      0: None; 1: version number with 
-			      Terre-con-erre-cigarro.maradns.org.
-			      2: 1 & number of threads currently running
-			        with Tnumthreads.maradns.org. */
+                              0: None; 1: version number with
+                              Terre-con-erre-cigarro.maradns.org.
+                              2: 1 & number of threads currently running
+                                with Tnumthreads.maradns.org. */
         "default_rrany_set", /* What kind of records do we return when someone
                                 sends us an ANY query.  Note:  This only
                                 changes MaraDNS' behavior as an authoritative
@@ -68,7 +70,8 @@ char *keywords[KEYCOUNT] = {
         "maximum_cache_elements", /* Maximum number of elements in the cache */
         "maxprocs",     /* Maximum number of udp threads or tcp processes */
         "min_ttl_cname", /* Minimum TTL for CNAME records */
-	"min_ttl", /* Minimum TTL for records */
+        "min_ttl", /* Minimum TTL for authoritative records */
+        "min_visible_ttl", /* Minimum TTL that MaraDNS will report */
         "no_fingerprint", /* Do we remove certain distinctive traits that
                              MaraDNS has; e.g. a TXT query of
                              erre-con-erre-cigarro.maradns.org. tells us
@@ -84,28 +87,62 @@ char *keywords[KEYCOUNT] = {
                               DNS server to respond before we give up and
                               try the next DNS server. */
         "timestamp_type", /* People get religious about what is
-	                     the True Timestamp Format (TM); as a
-		     	     result we need to allow the user to 
-			     choose this or the mailing list will
-			     have a heated discussion about what 
-			     format to use every month or so */
-        "tuzona_inetd", /* Do we run the zoneserver from inetd (security risk:
-                           enabling this disables the zone_transfer_acl)
-                           This feature is deprecated (never was 
-                           documented) */
-        "win9x_service", /* Run as Win9x (not NT) service (Thanks JEB) */
+                             the True Timestamp Format (TM); as a
+                             result we need to allow the user to
+                             choose this or the mailing list will
+                             have a heated discussion about what
+                             format to use every month or so */
         "verbose_level", /* How verbose error messages are */
-	"tcp_convert_acl", /* Who is allowed to use the zoneserver to forward 
-	                      requests to a UDP DNS server */
+        "tcp_convert_acl", /* Who is allowed to use the zoneserver to forward
+                              requests to a UDP DNS server */
         "tcp_convert_server", /* The ip of the UDP dns server or zoneserver
-	                         forwards requests to */
+                                 forwards requests to */
         "tcp_convert_recursion", /* Whether or not to ask for recursion when
-	                            sending requests to the UDP DNS server */
-        "resurrections", /* Whether it is possible to look at expired 
-                            data in case any of the nameservers for a 
-                            given record can not be contacted; named
-                            thusly because we "ressurect" expired data
-                            if we need to */
+                                    sending requests to the UDP DNS server */
+        "reject_aaaa", /* Send a bogus SOA whenever an AAAA request is sent
+                          to the server */
+        "reject_ptr", /* Send a bogus SOA whenever a PTR request is sent
+                         to the server */
+        "verbose_query", /* Whether to output every query the recursive DNS
+                            server makes */
+        "csv2_default_zonefile", /* Default ("*" character allowed at end of
+                                    hostname) zonefile */
+        "dos_protection_level", /* How much stuff we disable in MaraDNS
+                                   to protect her from a DOS attack */
+        "no_cname_warnings", /* Whether to supress warnings about dangling
+                                CNAME records */
+        "long_packet_ipv4", /* The ipv4 addresses that we send long packets
+                               to */
+        "synth_soa_origin", /* The origin to put in a synthetic SOA record */
+        "synth_soa_serial", /* The format for the serial in a synthetic SOA
+                               record */
+        "bind_star_handling", /* Whether or not to handle star records the
+                                 way BIND does */
+        "admin_acl", /* List of IPs allowed to administrate MaraDNS */
+        "remote_admin", /* Whether verbose_level can be remotely
+                         * set while MaraDNS is running */
+        "handle_noreply", /* How to handle the case of none of the remote
+                           * servers replying at all */
+        "retry_cycles", /* The number of times we try to contact all of
+                         * the remote DNS servers to resolve a given name */
+        "csv2_tilde_handling", /* How to handle tildes in CSV2 zone files */
+        "dns_port", /* What port to bind MaraDNS to */
+        "upstream_port", /* What port to contact when contacting other
+                            DNS servers */
+        "recurse_delegation", /* They might just want to recurse in the
+                               * case when the server would otherwise give
+                               * out a delegation NS record.  This is by
+                               * default disabled, since turning this on
+                               * confuses people */
+        "recurse_min_bind_port", /* The lowest numbered port the recursive
+                                  * resolver will bind to */
+        "recurse_number_ports", /* The number of ports the recursive resolver
+                                 * is allowed to bind to */
+        "max_mem", /* The maximum amount of memory we allow MaraDNS to
+                    * allocate, in kilobytes */
+        "notthere_ip", /* The IP to give users when they try to recursively
+                        * query a host that is not there or isn't responding
+                        */
         "zone_transfer_acl" /* ACL of IPs allowed to perform zone transfers */
         };
 
@@ -122,6 +159,27 @@ int mara_goodjs(js_string *test) {
     return JS_SUCCESS;
     }
 
+/* is_numeric_js: Determine if a given js_string object is really
+ * a number.  0 if it isn't, 1 if it is, JS_ERROR on error */
+
+int is_numeric_js(js_string *test) {
+        int counter;
+        if(!mara_goodjs(test)) {
+                return JS_ERROR;
+        }
+        if(test->unit_count > test->max_count) {
+                return JS_ERROR;
+        }
+        if(test->unit_count == 0)
+                return 0;
+        for(counter = 0; counter < test->unit_count; counter++) {
+                if(*(test->string + counter) < '0' ||
+                                *(test->string + counter) > '9')
+                        return 0;
+        }
+        return 1;
+}
+
 js_string *kvar[KEYCOUNT];
 
 /* keyword2num: Convert a keyword (like "kiwi_maillog") to a number
@@ -149,7 +207,7 @@ int keyword2num(js_string *keyword) {
             return JS_ERROR;
             }
         js_set_encode(name,MARA_LOCALE);
-	/* HACK: We don't use encode, so just make both the same */
+        /* HACK: We don't use encode, so just make both the same */
         js_set_encode(keyword,MARA_LOCALE);
         if(js_issame(keyword,name)) {
             js_destroy(name);
@@ -198,7 +256,7 @@ int init_kvars() {
    input: name, value
    output: JS_SUCCESS if set, JS_ERROR if fatal error, or
            0 if the variable has not been set in the user's mararc
-	   (or is zero-length)
+           (or is zero-length)
    global vars used: kvar[], keywords[]
 */
 
@@ -239,7 +297,7 @@ int read_kvar(js_string *name, js_string *value) {
    output: JS_SUCCESS or JS_ERROR, depending on error/success
    global vars used: kvar[], keywords[]
 */
-int write_kvar(js_string *name, js_string *value) {
+int write_kvar(js_string *name, js_string *value, int is_plus) {
 
     int num;
     num = keyword2num(name);
@@ -257,8 +315,21 @@ int write_kvar(js_string *name, js_string *value) {
         return num;
 
     /* Copy over the value */
-    if((js_copy(value,kvar[num])) == JS_ERROR)
-        return JS_ERROR;
+    if(is_plus != 1) {
+        if(js_copy(value,kvar[num]) == JS_ERROR)
+                return JS_ERROR;
+    } else { /* += parsing */
+            /* Make sure it is there; this is not perfect */
+            if(is_numeric_js(kvar[num]) == 1) {
+                    return -5; /* Sorry, += only supported for strings */
+            }
+            if(js_length(kvar[num]) == 0) {
+                    return -4; /* One needs to declare += string first */
+            }
+            if(js_append(value,kvar[num]) == JS_ERROR) {
+                    return JS_ERROR;
+            }
+    }
 
     return JS_SUCCESS;
 
@@ -266,13 +337,14 @@ int write_kvar(js_string *name, js_string *value) {
 
 /* Keywords that are dictionaries in the MaraDNS rc file */
 
-#define DKEYCOUNT 4
+#define DKEYCOUNT 6
 
 char *dkeywords[DKEYCOUNT] = {
-        "csv1",
+        "csv1", "csv2",
         "root_servers",
-	"upstream_servers",
-        "ipv4_alias" };
+        "upstream_servers",
+        "ipv4_alias",
+        "future" };
 
 mhash *dvar[DKEYCOUNT];
 
@@ -287,7 +359,23 @@ mhash *dvar_raw(int index) {
     return dvar[index];
     }
 
-/* dkeyword2num: Convert a keyword (like "csv") to a number
+/* dq_keyword2n: Convert a null-terminated string (like "csv1")
+ * to a number (0, in this case)
+ * input: A null-terminated string with the keyword
+ * Output: The number of the keyword (starting at 0), JS_ERROR on error,
+ * -2 on no match
+ */
+
+int dq_keyword2n(char *in) {
+        int counter;
+        for(counter = 0; counter < DKEYCOUNT; counter++) {
+                if(!strncmp(in,dkeywords[counter],128))
+                        return counter;
+        }
+        return -2;
+}
+
+/* dkeyword2num: Convert a keyword (like "csv2") to a number
                 (1, in this case)
    input: A js_string object with the keyword
    output: The number of the keyword (starting at 0), JS_ERROR on error,
@@ -349,7 +437,7 @@ int init_dvars() {
     for(counter = 0;counter < DKEYCOUNT;counter++)
         dvar[counter] = 0;
 
-    return JS_SUCCESS; 
+    return JS_SUCCESS;
     }
 
 /* new_dvar: Create a new mhash object for a given dvar object.
@@ -399,9 +487,9 @@ int read_dvar(js_string *name, js_string *key, js_string *value) {
     /* Find out where the actual value is */
     place = mhash_get_js(dvar[num],key);
     /* This is a little klunky; there really needs to be a special
-       "yes, the hash is there, but I can't find the element you are 
+       "yes, the hash is there, but I can't find the element you are
         seeking" error code.  However, other MaraDNS code depends on
-	this buggy behavior, so I am not changing this quite yet */
+        this buggy behavior, so I am not changing this quite yet */
     if(place == 0)
         return JS_ERROR;
 
@@ -419,10 +507,13 @@ int read_dvar(js_string *name, js_string *key, js_string *value) {
    output: JS_ERROR on error, 1 on most sucessful adds, 2 if the add
            caused the hash table to grow, -2 if the hash was not found
            (unused in mararc file), -3 if there is already a hash element
-           there
+           there, -4 for "Can't do += to something not there",
+           -5 for "Sorry, += only supported for strings",
+           -6 if dictionary has not been initialized
    global vars used: dvar[], keywords[]
 */
-int write_dvar(js_string *name, js_string *key, js_string *value) {
+int write_dvar(js_string *name, js_string *key, js_string *value,
+                int is_plus) {
 
     int num;
 
@@ -437,20 +528,83 @@ int write_dvar(js_string *name, js_string *key, js_string *value) {
         return JS_ERROR;
     if(num == JS_ERROR)
         return JS_ERROR;
-    if(dvar[num] == 0 || num < 0 || num > DKEYCOUNT)
+    if(dvar[num] == 0)
+        return -6;
+    if(num < 0 || num > DKEYCOUNT)
         return JS_ERROR;
 
     /* Return not found if not found */
     if(num == -2)
         return num;
 
-    /* If it already exists, return a -3 */
-    if(mhash_get_js(dvar[num],key) != 0)
+    /* If num is 0 (csv1), 1 (csv2), 2 (root_servers), or 3 (upstream_servers),
+     * then we make the key all lower-case */
+    if(num <= 3) {
+        int y,z;
+        y = 0;
+        for(z = 0; z < key->unit_count; z++) {
+            if(*(key->string + z) <= 'Z' && *(key->string + z) >= 'A') {
+                y = 1;
+                *(key->string + z) += 32;
+                }
+            }
+            if(y == 1) {
+                if(num <= 1) {
+                    printf("csv%d zone name ",num + 1);
+                } else if(num == 2) {
+                    printf("root_servers name ");
+                } else if(num == 3) {
+                    printf("upstream_servers name ");
+                } else {
+                    printf("unknown num %d name ",num);
+                }
+                show_esc_stdout(key);
+                printf(" had upper-case letters.  Converted.\n");
+                }
+        }
+
+    /* If it already exists, return a -3 (unless using += operator) */
+    if(mhash_get_js(dvar[num],key) != 0 && is_plus != 1)
         return -3;
 
     /* Copy over the value */
-    if((mhash_put_js(dvar[num],key,value)) == JS_ERROR)
-        return JS_ERROR;
+    if(is_plus != 1) {
+        if((mhash_put_js(dvar[num],key,value)) == JS_ERROR)
+                return JS_ERROR;
+    } else { /* += parsing */
+            js_string *see, *base;
+            if((base = mhash_get_js(dvar[num],key)) == 0) {
+                    return -4; /* Can't do += to something not there */
+            }
+            if(is_numeric_js(base) == 1) {
+                    return -5; /* Sorry, += only supported for strings */
+            }
+            if((see = js_create(256,1)) == 0) {
+                    return JS_ERROR;
+            }
+            if(js_set_encode(see,MARA_LOCALE) == JS_ERROR) {
+                    return JS_ERROR;
+            }
+            if(js_copy(base,see) == JS_ERROR) {
+                    return JS_ERROR;
+            }
+            if(js_append(value,see) == JS_ERROR) {
+                    js_destroy(see);
+                    return JS_ERROR;
+            }
+            if(js_copy(see,value) == JS_ERROR) {
+                    js_destroy(see);
+                    return JS_ERROR;
+            }
+            js_destroy(see);
+            if((see = (js_string *)mhash_undef(dvar[num],key)) == 0) {
+                    return JS_ERROR;
+            }
+            js_destroy(see);
+            if((mhash_put_js(dvar[num],key,value)) == JS_ERROR) {
+                    return JS_ERROR;
+            }
+    }
 
     /* Grow the table if needed */
     return mhash_autogrow(dvar[num]);
@@ -466,15 +620,16 @@ int write_dvar(js_string *name, js_string *key, js_string *value) {
           A js_string object that will contain the variable's value
           (description of problem on syntax error)
    output: JS_ERROR on fatal error, 1 if it is a string (or syntax error),
-           2 if it is an assosciative array, 3 if it inits a dict
+           2 if it is an assosciative array (dictionary), 3 if it inits a
+           dictionary
    note:  This needs to be rewritten to be less hackish
 */
 
 int parseline(js_string *line, js_string *var, js_string *key,
-              js_string *value) {
+              js_string *value, int *do_plus) {
 
     static js_string *quotes = 0, *alphanumeric = 0, *numbers = 0,
-           *equals = 0, *hashq = 0, *bslashq = 0, *allq = 0,
+           *equals = 0, *plusq = 0, *hashq = 0, *bslashq = 0, *allq = 0,
            *leftq = 0, *rightq = 0, *cleftq = 0, *crightq = 0,
            *blankq = 0;
     int quote1, quote2, quote3, quote4; /* Location of quotes in
@@ -486,6 +641,7 @@ int parseline(js_string *line, js_string *var, js_string *key,
     int tempp; /* temporary pointer */
     int ret = JS_SUCCESS; /* return value */
     char quote = '"';
+    char plus = '+';
     char equal = '=';
     char hash = '#';
     char left = '[';
@@ -494,6 +650,9 @@ int parseline(js_string *line, js_string *var, js_string *key,
     char cright = '}';
     char bslash = '\\';
 
+    if(do_plus != 0)
+            *do_plus = 0;
+
     /* Sanity checks */
     if(mara_goodjs(line) == JS_ERROR)
         return JS_ERROR;
@@ -513,6 +672,10 @@ int parseline(js_string *line, js_string *var, js_string *key,
     if(alphanumeric == 0)
         if((alphanumeric = js_create(256,1)) == 0)
                 return JS_ERROR;
+    if(plusq == 0) {
+            if((plusq = js_create(256,1)) == 0)
+                    return JS_ERROR;
+    }
     if(equals == 0)
         if((equals = js_create(256,1)) == 0)
                 return JS_ERROR;
@@ -543,6 +706,7 @@ int parseline(js_string *line, js_string *var, js_string *key,
 
     /* Initialize the various sets we look for */
     js_str2js(equals,&equal,1,1);
+    js_str2js(plusq,&plus,1,1);
     js_str2js(hashq,&hash,1,1);
     js_str2js(bslashq,&bslash,1,1);
     js_str2js(leftq,&left,1,1);
@@ -567,6 +731,7 @@ int parseline(js_string *line, js_string *var, js_string *key,
     js_str2js(quotes,&quote,1,1);
     js_append(quotes,allq);
     js_append(equals,allq);
+    js_append(plusq,allq);
     js_append(hashq,allq);
     js_append(bslashq,allq);
     js_append(alphanumeric,allq);
@@ -665,6 +830,21 @@ int parseline(js_string *line, js_string *var, js_string *key,
     /* Locate the equals */
     if((equalp = js_match(equals,line)) == JS_ERROR)
         return JS_ERROR;
+    /* See if there is a plus immediately before the equals */
+    if((tempp = js_match(plusq,line)) == JS_ERROR)
+            return JS_ERROR;
+    if(tempp != -2 && tempp != equalp - 1) {
+            js_qstr2js(var,"ERROR");
+            js_qstr2js(value,"+ only allowed in += form");
+            return JS_SUCCESS;
+    }
+    if(tempp == equalp - 1 && do_plus != 0)
+            *do_plus = 1;
+    if(js_match_offset(plusq,line,equalp) != -2) {
+            js_qstr2js(var,"ERROR");
+            js_qstr2js(value,"+ only allowed once on line");
+            return JS_SUCCESS;
+    }
 
     /* Find the beginning and end of the first word */
     varstart = varend = -2;
@@ -734,11 +914,15 @@ int parseline(js_string *line, js_string *var, js_string *key,
         valend = quote2 - 1;
         /* We need to make sure there are no bare words between varend and
            the first quote, allowing 'foo bar = "baz"' to return an error */
-        if(ret != 2 && /* Always matches in key["foo"] = "bar" cases */
-           js_match_offset(alphanumeric,line,varend + 1) < quote1) {
-            js_qstr2js(var,"ERROR");
-            js_qstr2js(value,L_2ND_WORD); /* "Second bare word found before quotes" */
-            return JS_SUCCESS;
+        if(ret != 2) {
+            int q;
+            /* Always matches in key["foo"] = "bar" cases */
+            q = js_match_offset(alphanumeric,line,varend + 1);
+            if(q < quote1 && q != -2) {
+                js_qstr2js(var,"ERROR");
+                js_qstr2js(value,L_2ND_WORD); /* "Second bare word found before quotes" */
+                return JS_SUCCESS;
+                }
             }
         }
     /* If the value is unquoted, then look for bare numbers after = */
@@ -875,6 +1059,8 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
     int error = 0; /* Line error is found on */
     int linenum = 1, command;
     int tnum; /* temporary number */
+    int is_plus = 0; /* Whether the line in question is a += instead of an =
+                        operator */
 
     static js_file *file = 0;
 
@@ -927,7 +1113,7 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
     if(js_open_read(fileloc,file) == JS_ERROR) {
         *errorret = -1;
         js_qstr2js(errorstr,L_CANNOT_OPEN); /* "Could not open mararc file at " */
-	js_append(fileloc,errorstr);
+        js_append(fileloc,errorstr);
         return JS_SUCCESS;
         }
     while(!js_buf_eof(file)) {
@@ -936,12 +1122,13 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
             js_close(file);
             return JS_ERROR;
             }
-        if((command = parseline(line,var,key,value)) == JS_ERROR) {
+        if((command = parseline(line,var,key,value,&is_plus)) == JS_ERROR) {
             js_qstr2js(errorstr,L_PARSELINE_FATAL); /* "Fatal error calling parseline" */
             js_close(file);
             return JS_ERROR;
             }
-        if(command == 2 && !error) { /* Add element to dictionary object */
+        if(command == 2 && !error) { /* Add or append element to
+                                      * dictionary object */
             tnum = dkeyword2num(var);
             if(tnum == JS_ERROR) {
                 js_qstr2js(errorstr,L_KEYWORD2NUM_FATAL); /* "Fatal error calling keyword2num" */
@@ -960,7 +1147,9 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
                     }
                 }
             else if(!error) {
-                if(write_dvar(var,key,value)==-3) {
+                int result;
+                result = write_dvar(var,key,value,is_plus);
+                if(result == -3) {
                 /* Exit if dictionary element exists more than once */
                     printf("%s",L_DT_PMRC);
                     js_show_stdout(var);
@@ -974,12 +1163,61 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
                     printf("%s","\"] = \"");
                     js_show_stdout(value);
                     printf("%s","\"\n");
+                    printf("Error in line %d\n",linenum);
+                    exit(1);
+                    }
+                else if(result == -4) {
+                    printf("+= operator target must be previously ");
+                    printf("defined with = operator\n");
+                    printf("Error in line %d\n",linenum);
+                    exit(1);
+                    }
+                else if(result == -5) {
+                    printf("+= operator target not supported for ");
+                    printf("numeric values\n");
+                    printf("Error in line %d\n",linenum);
+                    exit(1);
+                    }
+                else if(result == -6) {
+                    printf("Trying to access unitizalized dictionary var.\n");
+                    printf("Try adding this line to the beginning of the ");
+                    printf("mararc file:\n\n");
+                    js_show_stdout(var);
+                    printf(" = {}\n");
+                    printf("Error in line %d\n",linenum);
                     exit(1);
                     }
                 }
             }
         else if(command == 1 && !error) { /* Add element to string object */
             tnum = keyword2num(var);
+            if(tnum == -2) {
+                /* If an error happened parsing the line... */
+                js_qstr2js(tstr,"ERROR");
+                if(js_issame(tstr,var)) {
+                    if(!error) {
+                        error = linenum;
+                        *errorret = error;
+                        js_copy(value,errorstr);
+                        }
+                    }
+                }
+            if(tnum == -2 && var->unit_count > 0 && !error) {
+                printf("FATAL ERROR: Unknown mararc variable ");
+                show_esc_stdout(var);
+                printf("\nPlease look for the uncommented string \"");
+                show_esc_stdout(var);
+                printf("\"\nin your mararc file and remove this line.\n");
+                printf("\nThe line this error is on looks like this:\n");
+                /* Hackish way of chopping final newline */
+                if(*(line->string + line->unit_count - 1) == '\n' &&
+                   line->unit_count > 0) {
+                        line->unit_count--;
+                   }
+                show_esc_stdout(line);
+                printf("\n");
+                exit(10);
+                }
             if(tnum == JS_ERROR) {
                 js_qstr2js(errorstr,L_KEYWORD2NUM_FATAL); /* "Fatal error calling keyword2num" */
                 js_close(file);
@@ -996,8 +1234,22 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
                         }
                     }
                 }
-            else if(!error)
-                write_kvar(var,value);
+            else if(!error) {
+                    int result;
+                    result = write_kvar(var,value,is_plus);
+                    if(result == -4) {
+                            printf("+= operator target must be previously ");
+                            printf("defined with = operator\n");
+                            printf("Error in line %d\n",linenum);
+                            exit(1);
+                    }
+                    else if(result == -5) {
+                            printf("+= operator target not supported for ");
+                            printf("numeric values\n");
+                            printf("Error in line %d\n",linenum);
+                            exit(1);
+                    }
+                }
             }
         else if(command == 3)
                 new_dvar(var);
@@ -1009,6 +1261,40 @@ int read_mararc(js_string *fileloc,js_string *errorstr,int *errorret) {
 
     js_close(file);
 
+    /* DEBUG; show them everything set in the MaraRC file */
+    /* Disaled by commenting out */
+    /*
+    printf("Normal string variables: \n");
+    for(linenum = 0; linenum < KEYCOUNT; linenum++) {
+            printf("\t%s = \"",keywords[linenum]);
+            show_esc_stdout(kvar[linenum]);
+            printf("\"\n");
+    }
+    printf("\nDictionary variables: \n");
+    for(linenum = 0; linenum < DKEYCOUNT; linenum++) {
+            mhash *dvh;
+            js_string *key;
+            key = js_create(256,1);
+            dvh = (mhash *)dvar_raw(linenum);
+            if(dvh == 0) {
+                    printf("\t%s is undefined\n",dkeywords[linenum]);
+            } else {
+                    if(mhash_firstkey(dvh,key) == 0) {
+                            printf("\t%s is empty\n",dkeywords[linenum]);
+                    } else {
+                            do {
+                                    printf("\t%s[\"",dkeywords[linenum]);
+                                    show_esc_stdout(key);
+                                    printf("\"] = \"");
+                                    show_esc_stdout(mhash_get_js(dvh,key));
+                                    printf("\"\n");
+                            } while(mhash_nextkey(dvh,key) != 0);
+                    }
+            }
+    }
+    printf("\n\n"); */
+    /* End DEBUG code to show everything read in */
+
     return JS_SUCCESS;
     }
 
