@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2007 Sam Trenholme
+/* Copyright (c) 2002-2014 Sam Trenholme
  *
  * TERMS
  *
@@ -176,6 +176,9 @@ int change_rtype(js_string *js, int newtype) {
 void init_rr(rr *data) {
      data->expire = data->ttl = data->authoritative = data->rr_type = 0;
      data->next = data->ip = 0;
+#ifdef IPV6
+     data->ip6 = 0;
+#endif
      data->query = data->data = 0;
      data->ptr = 0;
      data->seen = 0;
@@ -1139,7 +1142,8 @@ int add_an(mhash *bighash, js_string *error) {
                    the answers */
                 record = qdata.value;
                 do {
-                    if(answer_ip_query(qtype,record->data,a_query) == JS_ERROR)
+                    if(record != 0 &&
+                       answer_ip_query(qtype,record->data,a_query) == JS_ERROR)
                         continue;
                     if(change_rtype(a_query,RR_A) == JS_ERROR)
                         continue;
@@ -1159,6 +1163,18 @@ int add_an(mhash *bighash, js_string *error) {
                                 }
                             }
                         }
+#ifdef IPV6
+                    if(change_rtype(a_query,RR_AAAA) == JS_ERROR)
+                        continue;
+                    adata = mhash_get(bighash,a_query);
+                    /* If so, point to it */
+                    if(adata.datatype == MARA_DNSRR) {
+                        if(record != 0 && adata.value != 0) {
+                            record->ip6 = adata.value;
+                            }
+                        else { if(record != 0) {record->ip6 = 0;} }
+                        }
+#endif
                     } while ((record = record->next) != 0);
                 }
             /* If it is a CNAME query, proceed to find the assosciated
