@@ -1,4 +1,4 @@
-/* Copyright (c) 2002-2007 Sam Trenholme
+/* Copyright (c) 2002-2014 Sam Trenholme
  *
  * TERMS
  *
@@ -75,7 +75,7 @@ int main(int argc, char **argv) {
 
     unsigned char len[2];
 
-    if((argc < 3 && argc > 4) || argv[1] == 0 || argv[2] == 0) {
+    if(argv[1] == 0 || argv[2] == 0) {
         harderror(L_USAGE_FETCH); /* "Usage: getzone zone_name zone_server_IP" */
         }
     if(argc == 4 && argv[3] == 0) {
