@@ -1,4 +1,4 @@
-/* Copyright (c) 2002,2003 Sam Trenholme
+/* Copyright (c) 2002-2010 Sam Trenholme and Albert Lee
  *
  * TERMS
  *
@@ -16,7 +16,7 @@
  * fitness for purpose.
  */
 
-/* This is the core DNS server */
+/* This is the core TCP DNS server */
 
 /* Language specific labels */
 #include "zoneserver_en.h"
@@ -25,14 +25,16 @@
 
 #include "../libs/MaraHash.h"
 #include "../MaraDns.h"
+#include "../server/read_kvars.h"
 #include <stdio.h>
 #include <unistd.h>
+#include <fcntl.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
 #include <time.h>
-#ifdef __FreeBSD__
+#if defined(__FreeBSD__) || defined(__DragonFly__)
 #include <sys/time.h>
 #endif
 #include <sys/types.h>
@@ -54,21 +56,41 @@
 /* Function prototypes */
 #include "../dns/functions_dns.h"
 #include "../parse/functions_parse.h"
-extern int decomp_init();
+#include "../parse/Csv2_database.h"
+#include "../parse/Csv2_read.h"
+#include "../parse/Csv2_functions.h"
+#include "functions_tcp.h"
 
 /* One global variable: The number of children processes running */
 int num_children = 0;
 /* Another global variable: How many times they have run srng() before */
 int srng_place = 0;
+/* Yet another global variable: How csv2 zone files handle the tilde
+ * character */
+int csv2_tilde_handling = 2;
+/* And another global variable: How much logging we do */
+int verbose = 0;
+/* Placeholder; dns_records_served is used by MaraDNS.c */
+int dns_records_served = 0;
 /* The last global variable: The IP of the UDP server we forward queries
    to */
 int udp_forward_server = 0;
 
+int dns_port = 53; /* The default port for the zoneserver to listen on */
+
+int no_cname_warnings = 1; /* So we can link to MaraBigHash.o */
+
 /* Signal handler for handling the exit of a child */
 void handle_childs() {
     if(waitpid(0,NULL,WNOHANG) > 0)
         num_children--;
-    }
+}
+
+/* Signal handler for termination of the root process */
+void handle_term() {
+    killpg(getpgrp(), SIGTERM);
+    exit(0);
+}
 
 /* Print out log messages
    Input: Null-terminated string with the message to log
@@ -91,6 +113,7 @@ int mlog(char *logmessage) {
 
 void harderror(char *why) {
     printf("%s%s%s",L_FATAL,why,LF); /* "Fatal error: ", why, "\n" */
+    killpg(getpgrp(), SIGTERM); /* Don't leave orphaned children */
     exit(3);
     }
 
@@ -105,11 +128,10 @@ uint16 srng() {
     js_string *srng_filename;
     unsigned char prng_seed[34];
     int desc, counter, max;
-    pid_t process_id;
+    pid_t process_id = 0;
     char path[MAXPATHLEN + 2];
     uint16 ret;
-    static BYTE r_inBlock[17],r_outBlock[17],r_binSeed[17];
-    static BYTE r_seedMaterial[320]; /* We may not eventually need this */
+    static MARA_BYTE r_inBlock[17],r_outBlock[17],r_binSeed[17];
     static keyInstance r_seedInst;
     static cipherInstance r_cipherInst;
 
@@ -136,6 +158,15 @@ uint16 srng() {
                queries will raz the end user; we must do it this way
                so that old mararc files work with the new zoneserver */
             }
+
+        /* Default random_seed_file: /dev/urandom */
+        if(js_length(srng_filename) == 0) {
+                if(js_qstr2js(srng_filename,"/dev/urandom") != JS_SUCCESS) {
+                        printf("AIEEEE! Can not set default random_seed_file");
+                        exit(1);
+                        }
+                }
+
         /* OK, now start doing the RNG stuff */
         memset(r_inBlock,0,16);
         time((time_t *)&r_inBlock[0]); /* Change it every second */
@@ -184,7 +215,7 @@ uint16 srng() {
             process_id >>= 8;
             }
         /* Initialize the PRNG with the seed in question */
-        if(makeKey(&r_seedInst, DIR_ENCRYPT, 128, prng_seed) != 1) {
+        if(makeKey(&r_seedInst, DIR_ENCRYPT, 128, (char *)prng_seed) != 1) {
             printf("AIEEE! Not able to make key\n");
             exit(1);
             }
@@ -218,52 +249,59 @@ uint16 srng() {
     return 12;
     }
 
-/* Bind to TCP port 53.
+/* Bind to TCP dns_port.
    Input: pointer to socket to bind on, js_string with the dotted-decimal
           ip address to bind to
    Output: JS_ERROR on error, JS_SUCCESS on success
 */
 
-int tcpbind(int *sock, js_string *ip_ddip) {
+int tcpbind(int *sock, uint32 ip) {
     int len_inet; /* Length */
     struct sockaddr_in dns_tcp;
-    char ip[64];
+    int on = 1;
 
     /* Sanity checks */
     if(sock == 0)
         return JS_ERROR;
-    if(js_has_sanity(ip_ddip) == JS_ERROR)
-        return JS_ERROR;
-    if(ip_ddip->unit_size != 1)
+    if(ip == 0xffffffff)
         return JS_ERROR;
 
-    /* Create string to hold IP */
-    if(js_js2str(ip_ddip,ip,60) == JS_ERROR)
-        return JS_ERROR;
+    /* This will convert ip into the correct form */
+    ip = htonl(ip);
 
     /* Create a raw TCP socket */
     if((*sock = socket(PF_INET,SOCK_STREAM,0)) == -1) {
         return JS_ERROR;
         }
+    /* Allow the socket to be reused more quickly (makes testing
+     * easier) */
+    if(setsockopt(*sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on,
+                  sizeof(on)) == -1) {
+        return JS_ERROR;
+        }
 
     /* Choose an IP and port to bind to */
     memset(&dns_tcp,0,sizeof(dns_tcp));
     dns_tcp.sin_family = AF_INET;
-    dns_tcp.sin_port = htons(53);
-    if((dns_tcp.sin_addr.s_addr = inet_addr(ip)) == INADDR_NONE)
+    dns_tcp.sin_port = htons(dns_port);
+    if((dns_tcp.sin_addr.s_addr = ip) == INADDR_NONE)
         return JS_ERROR;
 
     len_inet = sizeof(dns_tcp);
 
     /* Bind to the socket.  Note that we usually have to be root to do this */
-    if(bind(*sock,(struct sockaddr *)&dns_tcp,len_inet) == -1)
+    if(bind(*sock,(struct sockaddr *)&dns_tcp,len_inet) == -1) {
+        printf("Unable to bind to to IP: ");
+        printf("%s",strerror(errno));
+        printf("\n");
         return JS_ERROR;
+    }
 
     /* Set up an active listen on the socket */
     if(listen(*sock,250) == -1)
         return JS_ERROR;
 
-    /* We are now on TCP port 53.  Leave */
+    /* We are now on TCP dns_port.  Leave */
     return JS_SUCCESS;
     }
 
@@ -287,7 +325,8 @@ int gettcp(int *sock, ipv4pair *acl1, ipv4pair *acl2, ipv4pair *acl3,
     uint32 ip;
 
     len_inet = sizeof(adr_clnt);
-    ret = accept(*sock, (struct sockaddr *)&adr_clnt,&len_inet);
+    ret = accept(*sock, (struct sockaddr *)&adr_clnt,
+                 (socklen_t *)&len_inet);
     if(ret == -1)
         return JS_ERROR;
 
@@ -325,6 +364,13 @@ int gettcp(int *sock, ipv4pair *acl1, ipv4pair *acl2, ipv4pair *acl3,
         /* OK, they were not on the ACL list.  Clise the connection and
            return an error */
         close(ret);
+        if(verbose >= 4) {
+                printf("Failed zone transfer attempt from IP %d.%d.%d.%d\n",
+                ip >> 24,
+                (ip >> 16) & 0xff,
+                (ip >> 8) & 0xff,
+                ip & 0xff);
+                }
         return JS_ERROR;
         }
     else {
@@ -348,7 +394,8 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
 
     /* Magic voodoo to write *NIX TCP code */
     struct sockaddr_in dns_udp, server;
-    int udp_socket, s, result;
+    /* int udp_socket; */
+    int s, result;
     fd_set rx_set;
     struct timeval timeout;
     int *ipv4_ip;
@@ -368,10 +415,10 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
 
     tcp_id = (*(packet->string) << 8) | *(packet->string + 1);
 
-    printf("DEBUG udp_ip is %p\n",udp_ip);
+    /*printf("DEBUG udp_ip is %p\n",udp_ip); */
     /* Send a UDP packet with the query in question */
     ipv4_ip = (int *)udp_ip;
-    printf("DEBUG ipv4_ip is %d\n",*ipv4_ip);
+    /*printf("DEBUG ipv4_ip is %d\n",*ipv4_ip);*/
     if(*ipv4_ip == 0xffffffff || *ipv4_ip == 0) { /* 255.255.255.255 and
                                                      0.0.0.0 */
         return JS_ERROR;
@@ -411,9 +458,15 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
     header.nscount = 0;
     header.arcount = 0;
     /* Make a beginning of a DNS query from that header */
+#ifdef AUTHONLY
+    if((outdata = js_create(12345,1)) == 0) {
+        return JS_ERROR;
+        }
+#else
     if((outdata = js_create(2500,1)) == 0) {
         return JS_ERROR;
         }
+#endif
     if(make_hdr(&header,outdata) == JS_ERROR) {
         js_destroy(outdata);
         return JS_ERROR;
@@ -446,7 +499,7 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
         js_destroy(outdata);
         return JS_ERROR;
         }
-    printf("DEBUG: About to send out UDP packet to %d...\n",*ipv4_ip);
+    /*printf("DEBUG: About to send out UDP packet to %d...\n",*ipv4_ip);*/
     result = sendto(s,outdata->string,outdata->unit_count,0,
                     (struct sockaddr *)&server,len_inet);
     if(result < 0) {
@@ -468,15 +521,16 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
         }
 
     /* Get reply from the DNS server */
-    if((indata = js_create(600,1)) == 0) {
+    if((indata = js_create(4100,1)) == 0) {
         return JS_ERROR;
         }
-    printf("DEBUG about to get UDP packet from remote server\n");
+    /*printf("DEBUG about to get UDP packet from remote server\n");*/
     result = recvfrom(s,indata->string,indata->max_count,0,
-                      (struct sockaddr *)&dns_udp,&len_inet);
+                      (struct sockaddr *)&dns_udp,
+                      (socklen_t *)&len_inet);
     if(result < 0) {
         close(s);
-        printf("DEBUG: Never got reply from UDP server\n");
+        /*printf("DEBUG: Never got reply from UDP server\n");*/
         js_destroy(outdata);
         js_destroy(indata);
         return JS_ERROR;
@@ -485,7 +539,7 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
     /* Now that we are done with the socket, close it */
     close(s);
 
-    printf("DEBUG got packet from remote server\n");
+    /*printf("DEBUG got packet from remote server\n");*/
 
     indata->unit_count = result;
     /* Make sure that the data is sane */
@@ -503,7 +557,7 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
          return JS_ERROR;
          }
 
-    printf("DEBUG tcp_id: %d udp_id: %d sid: %d\n",tcp_id,udp_id,sid);
+    /*printf("DEBUG tcp_id: %d udp_id: %d sid: %d\n",tcp_id,udp_id,sid);*/
 
     /* Make sure the reply has the ID number of the TCP request */
     *(outdata->string) = (tcp_id & 0xff00) >> 8;
@@ -524,7 +578,7 @@ int convert_query(int tcp_connect, void *udp_ip, int udp_ip_type,
         }
     /* close the connection after the forward query is sent */
     close(tcp_connect);
-    printf("DEBUG gave client reply, bye bye!\n");
+    /*printf("DEBUG gave client reply, bye bye!\n");*/
     return JS_SUCCESS;
     }
 
@@ -546,8 +600,12 @@ int serve_zone(int connect, int perms) {
     js_string *name, *data, *zone, *query, *response, *soa, *filename,
               *binzone;
     int is_soa = 1, soa_q = 0;
+    int counter;
     uint32 ttl;
 
+    /* 30-second idle timeout */
+    alarm(30);
+
     /* Create the js_string objects */
     if((name = js_create((MAX_RECORD_LENGTH * 2) + 3,1)) == 0)
         return JS_ERROR;
@@ -590,6 +648,7 @@ int serve_zone(int connect, int perms) {
     if(decompress_data(name,data) == JS_ERROR)
         goto clean_nd;
 
+    /* show_esc_stdout(name); printf("\n"); */ /*DEBUG*/
     /* Convert the header */
     if(read_hdr(data,&header) == JS_ERROR)
         goto clean_nd;
@@ -600,10 +659,10 @@ int serve_zone(int connect, int perms) {
         goto clean_nd;
         }
 
-    /* If they only have permission to convert the query, do so */
+    /* If they only have permission to convert (proxy via UDP) the query,
+       do so */
     if((perms & 1) != 1) {
         if((perms & 2) == 2) {
-            printf("DEBUG: OK, moving on to convert_query 3\n");
             convert_query(connect,&udp_forward_server,4,data,perms);
             return JS_SUCCESS;
             }
@@ -626,7 +685,6 @@ int serve_zone(int connect, int perms) {
     if(*(data->string + 12 + length) != 0) {
         /* We may need to forward the query to the UDP DNS upstream */
         if((perms & 2) == 2) {
-            printf("DEBUG: OK, moving on to convert_query 2\n");
             convert_query(connect,&udp_forward_server,4,data,perms);
             return JS_SUCCESS;
             }
@@ -636,8 +694,8 @@ int serve_zone(int connect, int perms) {
     switch(*(data->string + 13 + length)) {
         case RR_SOA: /* SOA */
             soa_q = 1;
-        case 251: /* IXFR */
-        case 252: /* AXFR */
+        case RR_IXFR: /* IXFR */
+        case RR_AXFR: /* AXFR */
             break;
         default:
             /* In the case of them having permission to both server
@@ -645,7 +703,6 @@ int serve_zone(int connect, int perms) {
                we need to forward the request when they ask for anything
                besides a SOA, IXFR, or AXFR */
             if((perms & 2) == 2) {
-                printf("DEBUG: OK, moving on to convert_query\n");
                 convert_query(connect,&udp_forward_server,4,data,perms);
                 return JS_SUCCESS;
                 }
@@ -661,6 +718,14 @@ int serve_zone(int connect, int perms) {
     if(js_copy(name,binzone) == JS_ERROR) {
         goto clean_ndb;
         }
+    /* Convert the name in to all lowercase (since A-Z never appear as a
+       metacharacter in a zone file, we can quickly do this conversion) */
+    for(counter = 0; counter < name->unit_count; counter++) {
+        if(*(name->string + counter) >= 'A' &&
+           *(name->string + counter) <= 'Z') {
+              *(name->string + counter) += 32;
+              }
+        }
     /* Convert the name in to a dotted decimal format that the mararc
        file uses */
     if(hname_translate(name,RR_A) == JS_ERROR)
@@ -671,10 +736,20 @@ int serve_zone(int connect, int perms) {
     /* See if we have a zone file for the zone they are requesting */
     if(js_qstr2js(name,"csv1") == JS_ERROR)
         goto clean_ndb;
+    /* At this point, we disable timeouts (to give time to transfer large
+     * zones over slow links) */
+    alarm(0);
     if(read_dvar(name,data,name) == JS_ERROR) { /* XXX Covers "no such zone" */
-        mlog(L_NO_ZONE_HERE); /* "Zone we do not have asked for, disconnecting" */
-        goto clean_ndb;
-        }
+            /* If we don't have a csv1 zone, perhaps we have a csv2 zone;
+             * check in csv2_parse_zone_zoneserver */
+            if(csv2_parse_zone_zoneserver(data,connect,soa_q,&header) ==
+                            JS_SUCCESS) {
+                    /* Success; close the connection, etc. */
+                    exit(0); /* fork() makes all that easy :) */
+            }
+            mlog(L_NO_ZONE_HERE); /* "Zone we do not have asked for, disconnecting" */
+            goto clean_ndb;
+    }
 
     /* name now has the file with the zone in question */
     /* Create the filename */
@@ -934,14 +1009,25 @@ int main(int argc, char **argv) {
     js_string *mararc_loc, *errors, *chrootn, *kvar_str, *maxpstr,
               *kvar_query, *bind_address, *incoming, *uncomp, *verbstr;
     unsigned char chroot_zt[255];
-    int errorn, uid, sock, maxprocs, verbose, counter, connection,
+    int errorn, sock, maxprocs, counter, connection,
         inetd = 0;
 #ifndef DARWIN
     struct rlimit rlim;
 #endif
-    pid_t pid;
+    pid_t pid = 0;
+    uid_t uid;
+    gid_t gid;
     ipv4pair zonetransfer_acl[512], tcpconvert_acl[512], recursive_acl[512];
     ipv4pair tcpconvert_servers[512];
+    int synth_soa_serial;
+    js_string *synth_soa_origin;
+
+    /* Kill children processes when we are signaled */
+    if(setpgid(0,0)) {
+        printf("%s",strerror(errno)); /* harderror() would kill the group which may not be correct yet */
+        return 3;
+    }
+    signal(SIGTERM,handle_term);
 
     /* Initialize the strings (allocate memory for them, etc.) */
     if((mararc_loc = js_create(256,1)) == 0)
@@ -1018,22 +1104,20 @@ on %s\nFor usage information, type in 'man maradns'" */
     /* Initialize the psudo-random-number-generator */
     srng();
 
-    /* Determine if we are standalone, or if we should run from inetd */
-    if(js_qstr2js(kvar_query,"tuzona_inetd") == JS_ERROR)
-        harderror(L_MAKE_KQ);
-    if(read_kvar(kvar_query,kvar_str) == JS_SUCCESS) {
-        inetd = js_atoi(kvar_str,0);
-        }
+    inetd = 0;
 
     /* Get in to a state of least privledge ASAP */
 
     /* Limit the maximum number of processes */
-    if(js_qstr2js(kvar_query,"maxprocs") == JS_ERROR)
-        harderror(L_MAKE_KQ); /* "Could not create kvar_query" */
-    if(read_kvar(kvar_query,maxpstr) == JS_ERROR)
-        harderror(L_MAXPROCS); /* "Problem getting maxprocs value.\nmaxprocs must be set before starting the MaraDNS server" */
-    if((maxprocs = js_atoi(maxpstr,0)) == 0)
-        harderror(L_MAXPROCS_NUM); /* "Problem converting maxprocs to a number\nThis must be a non-zero number" */
+    maxprocs = read_numeric_kvar("maxprocs",128);
+
+    /* Determine how csv2 zone files will handle the tilde character */
+    csv2_tilde_handling = read_numeric_kvar("csv2_tilde_handling",2);
+    if(csv2_tilde_handling < 0 || csv2_tilde_handling > 3) {
+        harderror("csv2_tilde_handling "
+                  "must have a value between 0 and 3");
+        exit(1);
+        }
 
     /* If we have both maxprocs and max_tcp_procs defined, the zone server
        will use max_tcp_procs. */
@@ -1068,56 +1152,148 @@ on %s\nFor usage information, type in 'man maradns'" */
             harderror(L_MAKE_KQ); /* "Could not create kvar_query" */
         if(read_kvar(kvar_query,chrootn) == JS_ERROR)
             harderror(L_CHROOT); /* "Problem getting chroot kvar.\nYou must have chroot_dir set if you start this as root" */
-        if(js_js2str(chrootn,chroot_zt,200) == JS_ERROR)
+        if(js_js2str(chrootn,(char *)chroot_zt,200) == JS_ERROR)
             harderror(L_CHROOT_NT); /* "Problem making chroot nt string.\nMake sure the chroot directory is 200 chars or less" */
-        if(chdir(chroot_zt) != 0)
+        if(chdir((char *)chroot_zt) != 0)
             harderror(L_NO_CHROOT); /* "Problem changing to chroot dir.\nMake sure chroot_dir points to a valid directory" */
-        if(chroot(chroot_zt) != 0)
+        if(chroot((char *)chroot_zt) != 0)
             harderror(L_CHROOT_ERROR);  /* "Problem changing the root directory." */
 
-        if(inetd != 1) /* We have no logs if run from inetd */
-            mlog(L_CHROOT_SUCCESS); /* "Root directory changed" */
 
-        /* Bind to port 53
+        /* Determine which uid and gid to use */
+        uid = read_numeric_kvar("maradns_uid",MARADNS_DEFAULT_UID);
+        gid = read_numeric_kvar("maradns_gid",MARADNS_DEFAULT_GID);
+        if(uid < 10)  /* Security check */
+              uid = MARADNS_DEFAULT_UID;
+
+        mlog(L_CHROOT_SUCCESS); /* "Root directory changed" */
+
+        /* Bind to dns_port
            To Do: use capset to give us privledged bind abilities without
                   needing to be root.
         */
-        if(inetd != 1) { /* If we are a standalone server */
-            if(js_qstr2js(kvar_query,"bind_address") == JS_ERROR)
-                harderror(L_MAKE_KQ); /* "Could not create kvar_query" */
-            if(read_kvar(kvar_query,bind_address) == JS_ERROR)
-                harderror(L_BIND_GET); /* "Problem getting chroot kvar.\nYou must have bind_address set to the IP maradns will listen on" */
-            if(tcpbind(&sock,bind_address) == JS_ERROR)
-                harderror(L_BIND); /* "Problem binding to port 53.\nMost likely, another process is already listening on port 53" */
 
-            mlog(L_SOCKET_SUCCESS);  /* "Socket opened on TCP port 53" */
+        /* Set the dns_port */
+        dns_port = read_numeric_kvar("dns_port",53);
+        if(dns_port < 1 || dns_port > 65530) {
+            harderror("dns_port must be between 1 and 65530");
+            exit(1);
+        }
+
+        if(inetd != 1) { /* If we are a standalone server */
+            ipv4pair *bind_addresses;
+            int bind_address_iterate;
+            int stream1[2]; /* Used for piping */
+            /* pid_t child[70]; */ /* Children's process IDs */
+
+            /* bind_address is now a js_string object in the
+             * form "10.1.2.3,10.1.2.4".  Use the make_ip_acl code to
+             * make a list of bind addresses from the bind_address string,
+             * then fork once for each bind address.  This will
+             * be somewhat like the code below that gets the zone transfer
+             * ACL list
+             * */
+
+            bind_addresses = libtcp_bind_address(1);
+            if(bind_addresses == 0 || bind_addresses[0].ip == 0xffffffff) {
+                    harderror("Could not make list of bind addresses");
+            }
+            bind_address_iterate = 0;
+            while(bind_address_iterate < 70 &&
+                    bind_addresses[bind_address_iterate].ip != 0xffffffff) {
+                    /* Spawn a child process that is only bound to this
+                     * IP */
+                    /* We set things up so the
+                     * child's stdout/stderr goes to the parent's
+                     * stdout/stderr */
+                    if(pipe(stream1) != 0)
+                            harderror("Pipe()'s broken");
+                    /* if((child[bind_address_iterate] = fork())) { * Parent */
+                    if((pid = fork())) { /* Parent or error */
+                    if(pid < 0)
+                        harderror("Could not fork");
+                            close(stream1[1]);
+                            fcntl(stream1[0],F_SETFL,O_NONBLOCK);
+                            /* The following might not be portable */
+                            if(stream1[0] != bind_address_iterate + 3)
+                                dup2(stream1[0],
+                               bind_address_iterate + 3); /* Pipe redirected */
+                    } else { /* Child */
+                            close(stream1[0]);
+                            dup2(stream1[1],1); /* Stdout redirected to pipe */
+                            dup2(stream1[1],2); /* Stderr redirection */
+                            if(tcpbind(&sock,
+                        bind_addresses[bind_address_iterate].ip) == JS_ERROR)
+                            harderror(L_BIND); /* "Problem binding to dns_port.\nMost likely, another process is already listening on dns_port" */
+                            break;
+                    }
+                    bind_address_iterate++;
+            }
+            if(pid) { /* Parent */
+                    /* Handle children's output */
+                    fd_set child_set;
+                    fd_set read_set;
+                    char buf[1024];
+                    ssize_t readed;
+                    setgid(gid);
+                    if(setuid(uid) != 0) {
+                            harderror("Parent couldn't drop UID\n");
+                    }
+                    FD_ZERO(&child_set);
+                    /* Add all child fds */
+                    for(counter = 3;
+                      counter < bind_address_iterate + 3; counter++) {
+                        FD_SET(counter,&child_set);
+                    }
+                    for(;;) {
+                            FD_ZERO(&read_set);
+                            /* select on available children */
+                            for(counter = 3;
+                              counter < bind_address_iterate + 3; counter++) {
+                                if(FD_ISSET(counter,&child_set))
+                                    FD_SET(counter,&read_set);
+                            }
+                            if(select(bind_address_iterate + 3,&read_set,
+                              NULL,NULL,NULL) > 0) {
+                                for(counter = 3;
+                                  counter < bind_address_iterate + 3; counter++) {
+                                    if(FD_ISSET(counter,&read_set)) {
+                                        readed = read(counter,buf,1024);
+                                        if(readed > 0) {
+                                            int nowarn;
+                                            nowarn = write(1,buf,readed);
+                                        } else {
+                                            /* No longer exists  */
+                                            close(counter);
+                                            FD_CLR(counter,&child_set);
+                                        }
+                                    }
+                                }
+                            } else {
+                                break;
+                            }
+                            waitpid(0,NULL,WNOHANG); /* Catch exited children */
+                    }
+                    return 0;
+            }
+            if(libtcp_create_bind_addrs() == JS_ERROR)
+                harderror("libtcp_create_synthip_addrs");
+            mlog(L_SOCKET_SUCCESS);  /* "Socket opened on TCP dns_port" */
             }
 
-        /* Drop the elevated privledges */
-        if(js_qstr2js(kvar_query,"maradns_uid") == JS_ERROR)
-            harderror(L_MAKE_KQ); /* "Could not create kvar_query" */
-        if(read_kvar(kvar_query,kvar_str) == JS_ERROR)
-            harderror(L_NO_UID); /* "Problem getting maradns_uid kvar.\nYou must have maradns_uid set if you start this as root" */
-        if((uid = js_atoi(kvar_str,0)) < 10)
-            harderror(L_UID_INVALID); /* "maradns_uid is less than 10 or not a number.\nThis uid must have a value of 10 or more" */
+        /* Drop the elevated privileges */
+        setgid(gid);
         if(setuid(uid) != 0)
             harderror(L_NODROP); /* "Could not drop root uid" */
         if(setuid(0) == 0)
             harderror(L_STILL_ROOT);  /* "We seem to still be root" */
 
         if(inetd != 1) /* If we are not called from inetd */
-            mlog(L_DROP_SUCCESS); /* "Root privledges dropped" */
+            mlog(L_DROP_SUCCESS); /* "Root privileges dropped" */
 
         }
-    else if(inetd != 1) {
-        /* Bind to port 53 as a non-root user */
-        if(js_qstr2js(kvar_query,"bind_address") == JS_ERROR)
-            harderror(L_MAKE_KQ); /* "Could not create kvar_query" */
-        if(read_kvar(kvar_query,bind_address) == JS_ERROR)
-            harderror(L_BIND_GET); /* "Problem getting chroot kvar.\nYou must have bind_address set to the IP maradns will listen on" */
-        if(tcpbind(&sock,bind_address) == JS_ERROR)
-            harderror(L_BE_ROOT); /* "Problem binding to port 53.\nYou should run this as root" */
-        mlog(L_SOCKET_SUCCESS);  /* "Socket opened on TCP port 53" */
+    else {
+        harderror("inetd is not zero; this is a fatal error.  Complain loudly on the MaraDNS\nmailing list.\n");
         }
 
     /* Make a database of IPs permitted to transfer zone file */
@@ -1129,12 +1305,13 @@ on %s\nFor usage information, type in 'man maradns'" */
     if(read_kvar(kvar_query,kvar_str) == JS_ERROR)
         harderror(L_NO_ACL); /* "Could not read zone_transfer_acl data" */
     if(inetd == 1) {
-        if(kvar_str->unit_count != 0)
-            harderror(L_ACL_INETD); /* "zone_transfer_acl must not be set when running in inetd mode" */
+        harderror("inetd is set.  This is a fatal error.  You should not be seeing this\nComplain loudly on the MaraDNS mailing list.");
         }
     else {
-        if(make_ip_acl(kvar_str,zonetransfer_acl,500,0) == JS_ERROR)
-            harderror(L_ACL_LIST); /* "Could not make ip ACL list" */
+        if(make_ip_acl(kvar_str,zonetransfer_acl,500,0) == JS_ERROR) {
+            /* harderror(L_ACL_LIST); *//* "Could not make ip ACL list" */
+            zonetransfer_acl[0].ip=0xffffffff; /* No zone transfer ACL */
+            }
         }
 
     /* Make a database of IPs permitted to convert TCP queries into UDP
@@ -1199,14 +1376,45 @@ on %s\nFor usage information, type in 'man maradns'" */
         }
     udp_forward_server = tcpconvert_servers[0].ip;
 
+    /* Set the synth_soa_origin and synth_soa_serial as needed */
+    /* Get the values for the synthetic SOA serial and the synthetic SOA
+       origin (called MNAME in RFC1035) */
+    synth_soa_serial = read_numeric_kvar("synth_soa_serial",1);
+    if(synth_soa_serial < 1 || synth_soa_serial > 2) {
+        harderror("Fatal: synth_soa_serial must be 1 or 2\n");
+        }
+    set_soa_serial(synth_soa_serial);
+    verbstr = read_string_kvar("synth_soa_origin");
+    if(verbstr != 0 && js_length(verbstr) > 0) {
+        synth_soa_origin = js_create(256,1);
+        if(synth_soa_origin == 0) {
+            harderror("Fatal: can not create synth_soa_origin string");
+            }
+        if(js_qstr2js(synth_soa_origin,"Z") != JS_SUCCESS) {
+            harderror("Fatal: could not make synth_soa_origin string Z");
+            }
+        if(js_append(verbstr,synth_soa_origin) != JS_SUCCESS) {
+            harderror("Fatal: could not append to synth_soa_origin string");
+            }
+        /* We should see if the origin already has a dot on the end */
+        if(js_qappend(".",synth_soa_origin) != JS_SUCCESS) {
+            harderror("Fatal: could not append final dot to soa_synth_origin");
+            }
+        /* Now we make that raw data */
+        if(hname_2rfc1035(synth_soa_origin) <= 0) {
+            harderror("Fatal: Malformed synth_soa_origin value.\n"
+            "Please make sure that synth_soa_origin is a valid hostname\n"
+            "*without* a dot at the end.  For example:\n\n"
+            "\tsynth_soa_origin = \"example.com\"\n");
+            }
+        set_soa_origin(synth_soa_origin);
+        }
+
     /* Initialize decompression */
     decomp_init(0);
 
-    /* If we are being called from an external helper, such as inetd
-       or tcpserver */
     if(inetd == 1) {
-        serve_zone(1,1);
-        exit(0);
+        harderror("inetd is not zero.  This is a fatal error\nComplain loudly on the MaraDNS mailing list.\n");
         }
 
     /* Set up a signal handler so we can decrement the number of children
@@ -1227,10 +1435,9 @@ on %s\nFor usage information, type in 'man maradns'" */
           7: Both zone transfer and forward with recursion enabled
         */
         if(verbose >= 2)
-            mlog(L_WAITING); /* "Awaiting data on port 53" */
+            mlog(L_WAITING); /* "Awaiting data on dns_port" */
         connection = gettcp(&sock,zonetransfer_acl,tcpconvert_acl,
                      recursive_acl,500,&permissions);
-        printf("DEBUG: permissions %d\n",permissions);
         if(connection == JS_ERROR)
             continue;
         if(verbose >= 2)
@@ -1258,6 +1465,9 @@ on %s\nFor usage information, type in 'man maradns'" */
         /* Hackish way to clean up child processes without needlessly slowing
            the program */
         while(waitpid(0,NULL,WNOHANG) > 0) num_children--;
+
+        /* Make sure to close any lingering open connections */
+        close(connection);
         }
 
     }
