@@ -1,4 +1,4 @@
-/* Copyright (c) 2002,2003 Sam Trenholme
+/* Copyright (c) 2002-2005 Sam Trenholme
  *
  * TERMS
  *
@@ -20,7 +20,7 @@
 
 #include <stdio.h>
 
-main() {
+int main() {
         unsigned char a;
         while(!feof(stdin)) {
                 a = getc(stdin);
@@ -32,6 +32,7 @@ main() {
                         printf("%c",a);
 
                 }
+        return 0;
         }
 
 
